# __init__.py

'''Tools for analyzing and registering cortical meshes.'''

from freesurfer import (freesurfer_subject,
                        Hemisphere as FreeSurferHemisphere,
                        Subject    as FreeSurferSubject)
from cortex     import (CorticalMesh, mesh_smooth)
from vision     import (retinotopy_data, empirical_retinotopy_data, predicted_retinotopy_data,
                        register_retinotopy, retinotopy_anchors, retinotopy_model,
                        neighborhood_cortical_magnification)

# Version information...
__version__ = '0.2.15'

description = 'Integrate Python environment with FreeSurfer and perform mesh registration'


def reload_neuropythy():
    '''
    reload_neuropythy() reloads all of the modules of neuropythy and returns the reloaded
    neuropythy module.
    '''
    import sys
    mdls = ('neuropythy.immutable',
            'neuropythy.util.command',
            'neuropythy.util',
            'neuropythy.java',
            'neuropythy.geometry.util',
            'neuropythy.geometry.mesh',
            'neuropythy.geometry',
            'neuropythy.topology',
            'neuropythy.cortex',
            'neuropythy.freesurfer.subject',
            'neuropythy.freesurfer',
            'neuropythy.registration.core',
            'neuropythy.registration',
            'neuropythy.vision.models',
            'neuropythy.vision.retinotopy',
            'neuropythy.vision.cmag',
            'neuropythy.vision',
            'neuropythy.commands.surface_to_ribbon',
            'neuropythy.commands.benson14_retinotopy',
            'neuropythy.commands.register_retinotopy',
            'neuropythy.commands')
    for mdl in mdls:
        if mdl in sys.modules:
            reload(sys.modules[mdl])
    return reload(sys.modules['neuropythy'])



####################################################################################################
# neuropythy/freesurfer/__init__.py
# This file defines the FreeSurfer tools that are available as part of neuropythy.

import os
from .subject import (Subject, Hemisphere, 
                      cortex_to_ribbon_map, cortex_to_ribbon, cortex_to_ribbon_map_lines,
                      find_subject_path, subject_paths, add_subject_path)

# how to construct a freesurfer subject:
__freesurfer_subjects = {}
def freesurfer_subject(name):
    '''
    freesurfer_subject(name) yields a freesurfer Subject object for the subject with the given name.
    Subjects are cached and not reloaded.
    Note that subects returned by freesurfer_subject() are always persistent Immutable objects; this
    means that you must create a transient version of the subject to modify it via the member
    function sub.transient().
    '''
    subpath = find_subject_path(name)
    if subpath is None: return None
    fpath = '/' + os.path.relpath(subpath, '/')
    if fpath in __freesurfer_subjects:
        return __freesurfer_subjects[fpath]
    else:
        sub = Subject(subpath).persist()
        if isinstance(sub, Subject): __freesurfer_subjects[fpath] = sub
        return sub



####################################################################################################
# neuropythy/geometry/mesh.ph
# Tools for interpolating from meshes.
# By Noah C. Benson

import numpy as np
import scipy as sp
import scipy.spatial as space
import os, math
from pysistence import make_dict
from numpy.linalg import norm

from neuropythy.immutable import Immutable
from .util import (triangle_area, triangle_address, alignment_matrix_3D,
                   cartesian_to_barycentric_3D, cartesian_to_barycentric_2D,
                   barycentric_to_cartesian, point_in_triangle)

class Mesh(Immutable):
    '''
    A Mesh object represents a triangle mesh in either 2D or 3D space.
    '''

    @staticmethod
    def __calculate_triangle_centers(triangles, coords):
        coords = np.asarray(coords)
        coords = coords if coords.shape[0] < coords.shape[1] else coords.T
        triangles = np.asarray(triangles)
        triangles = triangles if triangles.shape[0] < triangles.shape[1] else triangles.T
        return (coords[:,triangles[0]] + coords[:,triangles[1]] + coords[:,triangles[2]]).T / 3

    @staticmethod
    def __calculate_triangle_normals(triangles, coords):
        coords = np.asarray(coords)
        coords = coords if coords.shape[0] < coords.shape[1] else coords.T
        triangles = np.asarray(triangles)
        triangles = triangles if triangles.shape[0] < triangles.shape[1] else triangles.T
        tmp = coords[:, triangles[0]]
        u01 = coords[:, triangles[1]] - tmp
        u02 = coords[:, triangles[2]] - tmp
        xp = np.cross(u01, u02, axisa=0, axisb=0)
        xpnorms = np.sqrt(np.sum(xp**2, axis=1))
        zero = np.isclose(xpnorms, 0)
        zero_idcs = np.where(zero)
        xp[zeros_idcs,:] = 0
        return xp / (xpnorms + zero)
        
    
    def __init__(self, triangles, coordinates):
        coordinates = coordinates if coordinates.shape[0] < 4 else coordinates.T
        triangles = triangles if triangles.shape[0] == 3 else triangles.T
        Immutable.__init__(
            self,
            {},
            {'coordinates': coordinates.T, 'triangles': triangles.T},
            {'triangle_centers': (('triangles','coordinates'),
                                  lambda t,x: Mesh.__calculate_triangle_centers(t, x)),
             'triangle_normals': (('triangle','coordinates'),
                                  lambda t,x: Mesh.__calculate_triangle_normals(t, x)),
             'triangle_hash':    (('triangle_centers',), lambda x: space.cKDTree(x)),
             'vertex_hash':      (('coordinates',), lambda x: space.cKDTree(x))})
    def __repr__(self):
        return 'Mesh(<%d triangles>, <%d vertices>)' % (self.triangles.shape[0],
                                                        self.coordinates.shape[0])


    # True if the point is in the triangle, otherwise False; tri_no is an index into the triangle
    def _point_in_triangle(self, tri_no, pt):
        tri = self.coordinates[self.triangles[tri_no]]
        return point_in_triangle(tri, pt)

    def _find_triangle_search(self, x, k=24, searched=set([])):
        # This gets called when a container triangle isn't found; the idea is that k should
        # gradually increase until we find the container triangle; if k passes the max, then
        # we give up and assume no triangle is the container
        if k >= 288: return None
        (d,near) = self.triangle_hash.query(x, k=k)
        near = [n for n in near if n not in searched]
        searched = searched.union(near)
        tri_no = next((kk for kk in near if self._point_in_triangle(kk, x)), None)
        return (tri_no if tri_no is not None
                else self._find_triangle_search(x, k=(2*k), searched=searched))
    
    def nearest_vertex(self, pt):
        '''
        mesh.nearest_vertex(pt) yields the id number of the nearest vertex in the given
        mesh to the given point pt. If pt is an (n x dims) matrix of points, an id is given
        for each column of pt.
        '''
        (d,near) = self.vertex_hash.query(pt, k=1)
        return near

    def point_in_plane(self, tri_no, pt):
        '''
        r.point_in_plane(id, pt) yields the distance from the plane of the id'th triangle in the
        registration r to the given pt and the point in that plane as a tuple (d, x).
        '''
        tx = self.coordinates[self.triangles[tri_no]]
        n = np.cross(tx[1] - tx[0], tx[2] - tx[0])
        n /= norm(n)
        d = np.dot(n, pt - tx[0])
        return (abs(d), pt - n*d)
    
    def nearest_data(self, pt, k=2, n_jobs=1):
        '''
        mesh.nearest_data(pt) yields a tuple (k, d, x) of the matrix x containing the point(s)
        nearest the given point(s) pt that is/are in the mesh; a vector d if the distances between
        the point(s) pt and x; and k, the face index/indices of the triangles containing the 
        point(s) in x.
        Note that this function and those of this class are made for spherical meshes and are not
        intended to work with other kinds of complex topologies; though they might work 
        heuristically.
        '''
        pt = np.asarray(pt, dtype=np.float32)
        if len(pt.shape) == 1:
            r = self.nearest_data([pt], k=k, n_jobs=n_jobs)[0];
            return (r[0][0], r[1][0], r[2][0])
        pt = pt if pt.shape[1] == 3 else pt.T
        (d, near) = self.triangle_hash.query(pt, k=k)
        ids = [tri_no if tri_no is not None else self._find_triangle_search(x, 2*k, set(near_i))
               for (x, near_i) in zip(pt, near)
               for tri_no in [next((k for k in near_i if self._point_in_triangle(k, x)), None)]]
        pips = [self.point_in_plane(i, p) if i is not None else (0, None)
                for (i,p) in zip(ids, pt)]
        return (np.asarray(ids),
                np.asarray([d[0] for d in pips]),
                np.asarray([d[1] for d in pips]))

    def nearest(self, pt, k=2, n_jobs=1):
        '''
        mesh.nearest(pt) yields the point in the given mesh nearest the given array of points pts.
        '''
        dat = self.nearest_data(pt)
        return dat[2]

    def distance(self, pt, k=2, n_jobs=1):
        '''
        mesh.distance(pt) yields the distance to the nearest point in the given mesh from the points
        in the given matrix pt.
        '''
        dat = self.nearest_data(pt)
        return dat[1]

    def container(self, pt, k=2, n_jobs=1):
        '''
        mesh.container(pt) yields the id number of the nearest triangle in the given
        mesh to the given point pt. If pt is an (n x dims) matrix of points, an id is given
        for each column of pt.
        '''
        pt = np.asarray(pt, dtype=np.float32)
        (d, near) = self.triangle_hash.query(pt, k=k) #n_jobs fails?
        if len(pt.shape) == 1:
            tri_no = next((kk for kk in near if self._point_in_triangle(kk, pt)), None)
            return (tri_no if tri_no is not None
                    else self._find_triangle_search(pt, k=(2*k), searched=set(near)))
        else:
            return [(tri_no if tri_no is not None
                     else self._find_triangle_search(x, k=(2*k), searched=set(near_i)))
                    for (x, near_i) in zip(pt, near)
                    for tri_no in [next((kk for kk in near_i if self._point_in_triangle(kk, x)),
                                        None)]]

    def interpolate(self, x, data, 
                    smoothing=1, mask=None, null=None, method='automatic', n_jobs=1,
                    container_ids=None):
        '''
        mesh.interpolate(x, data) yields a numpy array of the data interpolated from the given
        array, data, which must contain the same number of elements as there are points in the Mesh
        object mesh, to the coordinates in the given point matrix x. Note that if x is a vector
        instead of a matrix, then just one value is returned.
        
        The following options are accepted:
          * mask (default: None) indicates that the given True/False or 0/1 valued list/array should
            be used; any point whose nearest neighbor (see below) is in the given mask will, instead
            of an interpolated value, be set to the null value (see null option).
          * null (default: None) indicates the value that should be placed in the returned result if
            either a vertex does not lie in any triangle or a vertex is masked out via the mask
            option.
          * smoothing (default: 1) assuming that the method is 'interpolate' or 'automatic', this
            is the exponent used to smooth the interpolated surface; 1 is pure linear interpolation
            while 2 represents a slightly smoother version of this. Note that this is not an order
            of interpolation option.
          * method (default: 'automatic') specifies what method to use for interpolation. The only
            currently supported methods are 'automatic' or 'nearest'. The 'nearest' method does not
            actually perform a nearest-neighbor interpolation but rather assigns to a destination
            vertex the value of the source vertex whose veronoi-like polygon contains the
            destination vertex; note that the term 'veronoi-like' is used here because it uses the
            Voronoi diagram that corresponds to the triangle mesh and not the true delaunay
            triangulation. The 'automatic' checks every destination vertex and assigns it the
            'nearest' value if that value would not be a number, otherwise it interpolates linearly
            within the vertex's source triangle.
          * n_jobs (default: 1) is passed along to the cKDTree.query method, so may be set to an
            integer to specify how many processors to use, or may be -1 to specify all processors.
        '''
        x = np.asarray(x)
        if len(x.shape) == 2:
            x = x if x.shape[1] == 2 or x.shape[1] == 3 else x.T
            sz = 2
            dims = x.shape[1]
        elif len(x.shape) == 1:
            sz = 1
            dims = x.shape[0]
            x = np.asarray([x])
        else:
            raise ValueError('interpolation points must be a matrix or vector')
        if dims != self.coordinates.shape[1]:
            raise ValueError('interpolation points have wrong dimensionality for mesh')
        data = np.asarray(data)
        data_t = (data.shape[0] != self.coordinates.shape[0])
        if data_t: data = data.T
        # Okay, switch on method:
        if method == 'nearest':
            data = self._interpolate_nearest(x, data, mask, null, n_jobs)
        elif method == 'automatic':
            data = self._interpolate_linear(x, data, mask, null, smoothing, 12, n_jobs)
        data = np.asarray(data)
        return data.T if data_t else data

    # perform nearest-neighbor interpolation
    def _interpolate_nearest(self, x, data, mask, null, n_jobs):
        # lookup the neighbors...
        (d, nei) = self.vertex_hash.query(x, k=1) #n_jobs fails? version problem?
        if mask is None:
            return [data[i] for i in nei]
        else:
            return [data[i] if mask[i] == 1 else null for i in nei]
    # perform linear interpolation
    def _interpolate_linear(self, coords, data, mask, null, smoothing, check_no, n_jobs):
        # first, find the triangle containing each point...
        tris = self.triangles
        ## we only query the nearest check_no triangles; otherwise we don't fine the container
        containers = self.container(coords, k=check_no, n_jobs=n_jobs)
        # Okay, now we interpolate for each triangle
        nulls = [null for _ in range(data.shape[1])] if len(data.shape) > 1 else null
        if mask is None:
            return [(nulls if tri_no is None
                     else self._interpolate_triangle(x, data, tris[tri_no], smoothing, nulls))
                    for (x, tri_no) in zip(coords, containers)]
        else:
            return [(nulls if tri_no is None or any(mask[u] == 0 for u in tris[tri_no])
                     else self._interpolate_triangle(x, data, tris[tri_no], smoothing, nulls))
                    for (x, tri_no) in zip(coords, containers)]
    def _interpolate_triangle(self, x, data, tri_vertices, smoothing, nulls):
        # we'll want to project things down to 2 dimensions:
        if len(x) == 3:
            mtx = alignment_matrix_3D(x, [0,0,1])[0:2].T
            # Project out what we don't want
            corners = np.dot(self.coordinates[tri_vertices], mtx)
            x = np.asarray([0,0])
        else:
            corners = self.coordinates[tri_vertices]
        # get the mini-triangles' areas
        a_area = triangle_area(x, corners[1], corners[2]) ** smoothing
        b_area = triangle_area(x, corners[2], corners[0]) ** smoothing
        c_area = triangle_area(x, corners[0], corners[1]) ** smoothing
        tot = a_area + b_area + c_area
        # and do the interpolation:
        if np.isclose(tot, 0):
            return nulls
        else:
            if len(data.shape) == 1:
                return np.dot([a_area, b_area, c_area], data[tri_vertices]) / tot
            else:
                return np.dot([[a_area, b_area, c_area]], data[tri_vertices])[0] / tot

    def address(self, data):
        '''
        mesh.address(X) yields a dictionary containing the address or addresses of the point or
        points given in the vector or coordinate matrix X. Addresses specify a single unique 
        topological location on the mesh such that deformations of the mesh will address the same
        points differently. To convert a point from one mesh to another isomorphic mesh, you can
        address the point in the first mesh then unaddress it in the second mesh.
        '''
        # we have to have a topology and registration for this to work...
        data = np.asarray(data)
        if len(data.shape) == 1:
            face_id = self.container(data)
            if face_id is None: return None
            tx = self.coordinates[self.triangles[face_id]]
        else:
            data = data if data.shape[1] == 3 or data.shape[1] == 2 else data.T
            face_id = self.container(data)
            faces = self.triangles[face_id]
            tx = np.transpose(np.asarray([self.coordinates[f] for f in faces]), (0,1,2))
        bc = cartesian_to_barycentric_3D(tx, data) if self.coordinates.shape[1] == 3 else \
             cartesian_to_barycentric_2D(tx, data)
        return {'face_id': face_id, 'coordinates': bc}

    def unaddress(self, data):
        '''
        mesh.unaddress(A) yields a coordinate matrix that is the result of unaddressing the given
        address dictionary A in the given mesh. See also mesh.address.
        '''
        if not isinstance(data, dict):
            raise ValueError('address data must be a dictionary')
        if 'face_id' not in data: raise ValueError('address must contain face_id')
        if 'coordinates' not in data: raise ValueError('address must contain coordinates')
        face_id = data['face_id']
        coords = data['coordinates']
        if all(hasattr(x, '__iter__') for x in (face_id, coords)):
            faces = self.triangles[face_id].T
            tx = np.transpose(np.asarray([self.coordinates[f] for f in faces]), (0,2,1))
        else:
            tx = np.asarray(self.coordinates[self.triangles[face_id]])
        return barycentric_to_cartesian(tx, coords)





####################################################################################################
# command/__init__.py
# The main function, if neuropythy is invoked directly as command.
# By Noah C. Benson

import os, sys, math
import pysistence

from .register_retinotopy import (register_retinotopy_command, register_retinotopy_help)
from .benson14_retinotopy import (benson14_retinotopy_command, benson14_retinotopy_help)
from .surface_to_ribbon   import (surface_to_ribbon_command,   surface_to_ribbon_help)

# The commands that can be run by main:
commands = pysistence.make_dict(
    register_retinotopy=register_retinotopy_command,
    benson14_retinotopy=benson14_retinotopy_command,
    surface_to_ribbon=surface_to_ribbon_command)



####################################################################################################
# neuropythy/retinotopy/models.py
# Importing and interpreting of flat mesh models for registration.
# By Noah C. Benson

import numpy                as     np
import scipy                as     sp
import scipy.spatial        as     space
from   pysistence           import make_dict
from   numbers              import Number
import numpy.linalg, os, math, gzip

import neuropythy.geometry   as     geo
import neuropythy.freesurfer as     nfs
import neuropythy.cortex     as     ncx
from   neuropythy.immutable  import Immutable
from   neuropythy.java       import (java_link, serialize_numpy,
                                     to_java_doubles, to_java_ints, to_java_array)

class RetinotopyModel:
    '''
    RetinotopyModel is a class designed to be inherited by other models of retinotopy; any class
    that inherits from RetinotopyModel must implement the following methods to work properly with
    the registration system of the neuropythy library.
    '''
    def angle_to_cortex(self, theta, rho):
        '''
        model.angle_to_cortex(theta, rho) yields a (k x 2) matrix in which each row corresponds to
        an area map (e.g. V2 or MT) and the two columns represent x and y coordinates of the
        predicted location in a flattened cortical map at which one would find the visual angle
        values of theta (polar angle) and rho (eccentricity). Theta should be in units of degrees
        between 0 (upper vertical meridian) and 180 (lower vertical meridian).
        If theta and rhos are both vectors of length n, then the result is an (n x k x 2) matrix
        with one entry for each theta/rho pair.
        '''
        raise NotImplementedError(
            'Object with base class RetinotopyModel did not override angle_to_cortex')
    def cortex_to_angle(self, x, y):
        '''
        model.cortex_to_angle(x, y) yields a vector of (polar-angle, eccentricity, id) corresponding
        to the given (x,y) coordinates from a cortical map. The id that is returned is a positive
        integer corresponding to an ROI label (e.g., V1, MT).
        If x and y are vectors of length n then the return value is an (n x 3) matrix in which each
        row corresponds to one (x,y) pair.
        '''
        raise NotImplementedError(
            'Object with base class RetinotopyModel did not override cortex_to_angle')

# How we construct a Schira Model:
class SchiraModel(RetinotopyModel):
    '''
    The SchiraModel class is a class that inherits from RetinotopyModel and acts as a Python wrapper
    around the Java nben.neuroscience.SchiraModel class; it handles conversion from visual field 
    angle to cortical surface coordinates and vice versa for the Banded Double-Sech model proposed
    in the following paper:
    Schira MM, Tyler CW, Spehar B, Breakspear M (2010) Modeling Magnification and Anisotropy in the
    Primate Foveal Confluence. PLOS Comput Biol 6(1):e1000651. doi:10.1371/journal.pcbi.1000651.
    '''

    # These are the accepted arguments to the model:
    default_parameters = {
        'A': 1.05,
        'B': 90.0,
        'lambda': 0.4,
        'psi': 0.15,
        'scale': [21.0, 21.0],
        'shear': [[1.0, 0.0], [0.0, 1.0]],
        'center': [-6.0, 0.0],
        'v1size': 1.2,
        'v2size': 0.6,
        'v3size': 0.4,
        'hv4size': 0.9,
        'v3asize': 0.9}
    # This function checks the given arguments to see if they are okay:
    def __check_parameters(self, parameters):
        # we don't care if there are extra parameters; we just make sure the given parameters make
        # sense, then return the full set of parameters
        opts = {
            k: parameters[k] if k in parameters else v
            for (k,v) in SchiraModel.default_parameters.iteritems()}
        return opts

    # This class is immutable: don't change the params to change the model; 
    # don't change the java object!
    def __setattr__(self, name, val):
        raise ValueError('The SchiraModel class is immutable; its objects cannot be edited')

    def __init__(self, **opts):
        # start by getting the proper parameters
        params = self.__check_parameters(opts)
        # Now, do the translations that we need...
        if isinstance(params['scale'], Number) or np.issubdtype(type(params['scale']), np.float):
            params['scale'] = [params['scale'], params['scale']]
        if (isinstance(params['shear'], Number) or np.issubdtype(type(params['shear']), np.float)) \
           and params['shear'] == 0:
            params['shear'] = [[1, 0], [0, 1]]
        elif params['shear'][0][0] != 1 or params['shear'][1][1] != 1:
            raise RuntimeError('shear matrix [0,0] elements and [1,1] elements must be 1!')
        if ((isinstance(params['center'], Number)
             or np.issubdtype(type(params['center']), np.float))
            and params['center'] == 0):
            params['center'] = [0.0, 0.0]
        self.__dict__['parameters'] = make_dict(params)
        # Okay, let's construct the object...
        self.__dict__['_java_object'] = java_link().jvm.nben.neuroscience.SchiraModel(
            params['A'],
            params['B'],
            params['lambda'],
            params['psi'],
            params['v1size'],
            params['v2size'],
            params['v3size'],
            params['hv4size'],
            params['v3asize'],
            params['center'][0],
            params['center'][1],
            params['scale'][0],
            params['scale'][1],
            params['shear'][0][1],
            params['shear'][1][0])
    
    def angle_to_cortex(self, theta, rho):
        iterTheta = hasattr(theta, '__iter__')
        iterRho = hasattr(rho, '__iter__')
        jarr = None
        if iterTheta and iterRho:
            if len(theta) != len(rho):
                raise RuntimeError('Arguments theta and rho must be the same length!')
            jarr = self._java_object.angleToCortex(to_java_doubles(theta), to_java_doubles(rho))
        elif iterTheta:
            jarr = self._java_object.angleToCortex(to_java_doubles(theta),
                                                   to_java_doubles([rho for t in theta]))
        elif iterRho:
            jarr = self._java_object.angleToCortex(to_java_doubles([theta for r in rho]),
                                                   to_java_doubles(rho))
        else:
            return self._java_object.angleToCortex(theta, rho)
        return np.asarray([[c for c in r] for r in jarr])
    def cortex_to_angle(self, x, y):
        iterX = hasattr(x, '__iter__')
        iterY = hasattr(y, '__iter__')
        jarr = None
        if iterX and iterY:
            if len(x) != len(y):
                raise RuntimeError('Arguments x and y must be the same length!')
            jarr = self._java_object.cortexToAngle(to_java_doubles(x), to_java_doubles(y))
        elif iterX:
            jarr = self._java_object.cortexToAngle(to_java_doubles(x),
                                                   to_java_doubles([y for i in x]))
        elif iterY:
            jarr = self._java_object.cortexToAngle(to_java_doubles([x for i in y]),
                                                   to_java_doubles(y))
        else:
            return self._java_object.cortexToAngle(x, y)
        return np.asarray([[c for c in r] for r in jarr])
        

class RetinotopyMeshModel(RetinotopyModel):
    '''
    RetinotopyMeshModel is a class that represents a retinotopic map or set of retinotopic maps on
    the flattened 2D cortex.
    RetinotopyMeshModel(tris, coords, polar_angle, eccen, areas) yields a retinotopy mesh model
    object in which the given triangle and coordinate matrices form the mesh and the polar_angle,
    eccen, and areas give the appropriate data for each vertex in coords. Note that the areas
    parameter should be 0 on any boundary vertex and an integer labelling the roi for any other
    vertex.
    '''

    def __init__(self, triangles, coordinates, angles, eccens, area_ids, transform=None):
        triangles   = np.asarray(triangles)
        coordinates = np.asarray(coordinates)
        triangles   = triangles   if triangles.shape[1] == 3   else triangles.T
        coordinates = coordinates if coordinates.shape[1] == 2 else coordinates.T
        angles      = np.asarray(angles)
        eccens      = np.asarray(eccens)
        area_ids    = np.asarray(map(int, area_ids))
        # The forward model is the projection from cortical map -> visual angle
        self.forward = geo.Mesh(triangles, coordinates)
        # The inverse model is a set of meshes from visual field space to the cortical map
        xs = coordinates[:,0]
        ys = coordinates[:,1]
        zs = eccens * np.exp(1j * (90 - angles)/180*math.pi)
        coords = np.asarray([zs.real, zs.imag]).T
        self.inverse = {
            area: geo.Mesh(np.asarray(tris), coords)
            # Iterate over all the unique areas;
            for area in list(set(area_ids) - set([0]))
            # bind the triangles (0 area_ids indicate borders):
            for tris in [[t for t in triangles if (set(area_ids[t]) - set([0])) == set([area])]]}
        # Note the transform:
        self.transform = np.asarray(transform) if transform is not None else None
        self.itransform = numpy.linalg.inv(transform) if transform is not None else None
        # Save the data:
        self.data = {}
        self.data['x'] = xs
        self.data['y'] = ys
        self.data['polar_angle'] = angles
        self.data['eccentricity'] = eccens
        # we have to fix the area_ids to be the mean of their neighbors when on a boundary:
        boundaryNeis = {}
        for (b,inside) in [(b, set(inside))
                           for t in triangles
                           for (bound, inside) in [([i for i in t if area_ids[i] == 0],
                                                    [i for i in t if area_ids[i] != 0])]
                           if len(bound) > 0 and len(inside) > 0
                           for b in bound]:
            if b not in boundaryNeis: boundaryNeis[b] = inside
            else: boundaryNeis[b] |= inside
        for (b,neis) in boundaryNeis.iteritems():
            area_ids[b] = np.mean(area_ids[list(neis)])
        self.data['id'] = area_ids

    def cortex_to_angle(self, x, y):
        'See RetinotopyModel.cortex_to_angle.'
        if not hasattr(x, '__iter__'):
            return self.cortex_to_angle([x], [y])[0]
        # start by applying the transform to the points
        tx = self.itransform
        xy = np.asarray([x,y]).T if tx is None else np.dot(tx, [x,y,[1 for i in x]])[0:2].T
        # we only need to interpolate from the inverse mesh in this case
        interp_ae = self.forward.interpolate(
            xy,
            [self.data[tt] for tt in ['polar_angle', 'eccentricity']],
            method='automatic',
            null=np.nan)
        interp_id = self.forward.interpolate(
            xy,
            self.data['id'],
            method='nearest',
            null=np.nan)
        interp = np.asarray([interp_ae[0], interp_ae[1], interp_id])
        bad = np.where(np.isnan(np.prod(interp, axis=0)))[0]
        interp[:,bad] = 0.0
        return interp

    def angle_to_cortex(self, theta, rho):
        'See help(neuropythy.registration.RetinotopyModel.angle_to_cortex).'
        if not hasattr(theta, '__iter__'):
            return self.angle_to_cortex([theta], [rho])[0]
        theta = np.asarray(theta)
        rho = np.asarray(rho)
        zs = np.asarray(
            rho * np.exp([np.complex(z) for z in 1j * ((90.0 - theta)/180.0*math.pi)]),
            dtype=np.complex)
        coords = np.asarray([zs.real, zs.imag]).T
        # we step through each area in the forward model and return the appropriate values
        tx = self.transform
        xvals = self.data['x']
        yvals = self.data['y']
        res = np.asarray(
            [[self.inverse[area].interpolate(coords, xvals, smoothing=1),
              self.inverse[area].interpolate(coords, yvals, smoothing=1)]
             for area in map(int, sorted(list(set(self.data['id']))))
             if area != 0]
        ).transpose((2,0,1))
        if tx is not None:
            res = np.asarray(
                [[np.dot(tx, [xy[0], xy[1], 1])[0:2] if xy[0] is not None else [None, None]
                  for xy in ptdat]
                 for ptdat in res])
        # there's a chance that the coords are outside the triangle mesh; we want to make sure
        # that these get handled correctly...
        for (i,ptdat) in enumerate(res):
            for row in ptdat:
                if None in set(row.flatten()) and rho[i] > 86 and rho[i] <= 90:
                    # we try to get a fixed version by reducing rho slightly
                    res[i] = self.angle_to_cortex(theta[i], rho[i] - 0.5);
        return res

             
class RegisteredRetinotopyModel(RetinotopyModel):
    '''
    RegisteredRetinotopyModel is a class that represents a retinotopic map or set of retinotopic
    maps on the flattened 2D cortex OR on the 3D cortical surface, via a registration and set of
    map projection parameters.
    RegisteredRetinotopyModel(model, projection_params) yields a retinotopy
    mesh model object in which the given RetinotopyModel object model describes the 2D 
    retinotopy that is predicted for the vertices that result from a map projection, defined using
    the given projection_params dictionary, of the given registration. In other words, the resulting
    RegisteredRetinotopyModel will, when given a hemisphere object to which to apply the model,
    will look up the appropriate registration (by the name registration_name) make a map projection
    of using the given projection_params dictionary, and apply the model to the resulting
    coordinates.
    '''

    def __init__(self, model, **projection_params):
        '''
        RegisteredRetinotopyModel(model, <projection parameters...>) yields a
        retinotopy mesh model object in which the given RetinotopyModel object model describes
        the 2D retinotopy that is predicted for the vertices that result from a map projection,
        defined using the given projection_params dictionary, of the given registration. In other
        words, the resulting RegisteredRetinotopyModel will, when given a hemisphere object to
        which to apply the model, will look up the appropriate registration (by the name
        registration_name) make a map projection of using the given projection_params dictionary,
        and apply the model to the resulting coordinates.
        See also neuropythy.freesurfer's Hemisphere.projection_data method for details on the
        projection parameters;
        '''
        self.model = model
        if 'registration' not in projection_params:
            projection_params['registration'] = 'fsaverage_sym'
        if 'chirality' in projection_params:
            chirality = projection_params['chirality']
        elif 'hemi' in projection_params:
            chirality = projection_params['hemi']
        elif 'hemisphere' in projection_params:
            chirality = projection_params['hemisphere']
        else:
            chirality = None
        if chirality is None:
            self.projection_data = nfs.Hemisphere._make_projection(**projection_params)
        else:
            sub = nfs.freesurfer_subject(projection_params['registration'])
            hemi = sub.LH if chirality.upper() == 'LH' else sub.RH
            self.projection_data = hemi.projection_data(**projection_params)
        
    def cortex_to_angle(self, *args):
        '''
        The cortex_to_angle method of the RegisteredRetinotopyModel class is identical to that
        of the RetinotopyModel class, but the method may be given a map, mesh, or hemisphere, in
        which case the result is applied to the coordinates after the appropriate transformation (if
        any) is first applied.
        '''
        if len(args) == 1:
            if isinstance(args[0], ncx.CorticalMesh):
                if args[0].coordinates.shape[0] == 2:
                    X = args[0].coordinates
                    return self.model.cortex_to_angle(X[0], X[1])
                else:
                    m = self.projection_data['forward_function'](args[0])
                    res = np.zeros((3, args[0].coordinates.shape[1]))
                    c2a = np.asarray(self.cortex_to_angle(m.coordinates))
                    res[:, m.vertex_labels] = c2a if len(c2a) == len(res) else c2a.T
                    return res
            elif isinstance(args[0], nfs.Hemisphere):
                regname = self.projection_data['registration']
                if regname is None or regname == 'native':
                    regname = args[0].subject.id
                if regname not in args[0].topology.registrations:
                    raise ValueError('given hemisphere is not registered to ' + regname)
                else:
                    return self.cortex_to_angle(args[0].registration_mesh(regname))
            else:
                X = np.asarray(args[0])
                if len(X.shape) != 2:
                    raise ValueError('given coordinate matrix must be rectangular')
                X = X if X.shape[0] == 2 or X.shape[0] == 3 else X.T
                if X.shape[0] == 2:
                    return self.model.cortex_to_angle(X[0], X[1])
                elif X.shape[0] == 3:
                    Xp = self.projection_data['forward_function'](X)
                    return self.model.cortex_to_angle(Xp[0], Xp[1])
                else:
                    raise ValueError('coordinate matrix must be 2 or 3 dimensional')
        else:
            return self.model.cortex_to_angle(*args)
                    
    def angle_to_cortex(self, *args):
        '''
        The angle_to_cortex method of the RegisteredRetinotopyModel class is identical to that
        of the RetinotopyModel class, but the method may be given a map, mesh, or hemisphere, in
        which case the result is applied to the 'polar_angle' and 'eccentricity' properties.
        '''
        if len(args) == 1:
            if isinstance(args[0], CorticalMesh) or isinstance(args[0], nfs.Hemisphere):
                ang = ncx.retinotopy_data(args[0])
                ecc = ncx.retinotopy_data(args[0])
                return self.model.angle_to_cortex(ang, ecc)
            else:
                tr = np.asarray(args)
                if tr.shape[1] == 2: tr = tr.T
                elif tr.shape[0] != 2: raise ValueError('cannot interpret argument')
                return self.model.angle_to_cortex(tr[0], tr[1])
        else:
            return self.model.angle_to_cortex(*args)
                    

def load_fmm_model(filename, radius=math.pi/3.0, sphere_radius=100.0):
    '''
    load_fmm_model(filename) yields the fmm model indicated by the given file name. Fmm models are
    triangle meshes that define a field value at every vertex as well as the parameters of a
    projection to the 2D cortical surface via a registartion name and projection parameters.
    The following options may be given:
      * radius (default: pi/3) specifies the radius that should be assumed by the model; fmm models
        do not specify radius values in their projection parameters, but pi/3 is usually large for
        localized models.
      * sphere_radius (default: 100) specifies the radius of the sphere that should be assumed by
        the model. Note that in Freesurfer, spheres have a radius of 100.
    '''
    if not os.path.exists(filename):
        models_path = os.path.join(os.path.dirname(__file__), '..', 'lib', 'models')
        # we look for it in a number of ways:
        fname = next((fnm
                      for beg in ['', models_path] for end in ['', '.fmm', '.fmm.gz']
                      for fnm0 in [filename + end]
                      for fnm in [fnm0 if beg is None else os.path.join(beg, fnm0)]
                      if os.path.exists(fnm)),
                     None)
        if fname is None:
            raise ValueError('Given model/file name does not exist: %s' % filename)
        filename = fname
    if not os.path.isfile(filename):
        raise ValueError('Given filename (%s) is not a file!' % filename)
    gz = True if len(filename) > 3 and filename[-3:] == '.gz' else False
    lines = None
    with (gzip.open(filename, 'rb') if gz else open(filename, 'r')) as f:
        lines = f.read().split('\n')
    if len(lines) < 3 or lines[0] != 'Flat Mesh Model Version: 1.0':
        raise ValueError('Given file does not contain to a valid flat mesh model!')
    n = int(lines[1].split(':')[1].strip())
    m = int(lines[2].split(':')[1].strip())
    reg = lines[3].split(':')[1].strip()
    hemi = lines[4].split(':')[1].strip().upper()
    center = map(float, lines[5].split(':')[1].strip().split(','))
    onxaxis = map(float, lines[6].split(':')[1].strip().split(','))
    method = lines[7].split(':')[1].strip().lower()
    tx = np.asarray(
        [map(float, row.split(','))
         for row in lines[8].split(':')[1].strip(' \t[]').split(';')])
    crds = []
    for row in lines[9:(n+9)]:
        try:
            (left,right) = row.split(' :: ')
            crds.append(map(float, left.split(',')))
        except:
            print row
            raise
    crds = np.asarray([map(float, left.split(','))
                       for row in lines[9:(n+9)]
                       for (left,right) in [row.split(' :: ')]])
    vals = np.asarray([map(float, right.split(','))
                       for row in lines[9:(n+9)]
                       for (left,right) in [row.split(' :: ')]])
    tris = -1 + np.asarray(
        [map(int, row.split(','))
         for row in lines[(n+9):(n+m+9)]])
    mdl = RegisteredRetinotopyModel(
        RetinotopyMeshModel(tris, crds,
                            90 - 180/math.pi*vals[:,0], vals[:,1], vals[:,2],
                            transform=tx),
        registration=reg,
        center=center,
        center_right=onxaxis,
        method=method,
        radius=radius,
        sphere_radius=sphere_radius,
        chirality=hemi)
    return mdl



####################################################################################################
# neuropythy/immutable/__init__.py
# Simple class to inherit for immutable lazily-loading classes.
# By Noah C. Benson

import itertools
import collections
from pysistence import make_dict
import pysistence
import copy
import numpy as np

class Immutable(object):
    '''
    The Immutable class can be overloaded by any class that wishes to be an immutable lazily-loading
    class---Immutable facilitates that. It does so by overloading the standard dictionary functions
    and accepting a list of lazy loading instructions in its constructor. The constructor for the
    Immutable class expects three arguments plus any number of optional key-val assignments. The 
    optional key-val assignments are placed into the settable member 'options'. The first three 
    arguments are: Immutable(settable_vals, const_vals, lazy_vals).
      * settable_vals must be a dict of settable value names (strings) mapped to a function which
        takes as an argument the self object and the value being assigned and returns the value
        that should actually be assigned.
      * const_vals must be a dict of constant value names (strings) mapped to constant values; these
        will be constant member values of the object.
      * lazy_vals must be a dict of lazy value names (strings) mapped to a 2-tuple of (1) a tuple of
        argument names and (2) a function that accepts these arguments and yields the lazy value;
        the argument names may be any settable, const, or lazy value name.
    '''

    # Calculate lazy dependencies between values for this object
    def __lazy_deps(self):
        # we build up this dependency tree:
        dep_tree = dict()
        for (sname, sdata) in self._settable_vals.iteritems():
            deps = {sname}
            # sweep through lazy vals, eliminating those that depend on name; do this until none are
            # removed on a sweep
            go = True
            while go:
                tmp = [lnm for (lnm, (ldeps, lfn)) in self._lazy_vals.iteritems()
                           if lnm not in deps and any(dep in deps for dep in ldeps)]
                go = len(tmp) > 0
                deps.update(tmp)
            dep_tree[sname] = (deps - {sname})
        return dep_tree

    # This function will clear the lazily-evaluated members when a given value is changed
    def __update_values(self, name):
        for nm in self._lazy_deps[name]:
            if nm in self.__dict__:
                del self.__dict__[nm]

    # This is the most important function, given the encapsulation of this class:
    def __setattr__(self, name, val):
        if name in self._settable_vals:
            if self.__dict__['_persistent']:
                raise ValueError(('Immutable object is persistent and cannot be changed; use ' +
                                  'the method using or the method transient to update the ' +
                                  'value %s') % name)
            fn = self._settable_vals[name]
            fixed = fn(self, val)
            if isinstance(fixed, np.ndarray):
                if fixed.flags.writeable:
                    fixed = np.array(fixed)
                    fixed.flags.writeable = False
            self.__dict__[name] = fixed
            self.__update_values(name)
        elif name in self._lazy_vals:
            raise ValueError('The member %s is a lazy value and cannot be set' % name)
        elif name in self._const_vals:
            raise ValueError('The member %s is a constant value and cannot be set' % name)
        else:
            raise ValueError('Unrecognized Immutable member: %s' % name)

    # The getattr method makes sure that lazy members are computed when requested
    def __getattr__(self, name):
        if name[0] == '_' or name in self.__dict__ or name in Immutable.__dict__:
            return object.__getattribute__(self, name)
        elif name in self.__dict__['_lazy_vals']:
            (deps, fn) = self.__dict__['_lazy_vals'][name]
            tmp = fn(*[getattr(self, x) for x in deps])
            self.__dict__[name] = tmp
            return tmp
        elif name in self._const_vals:
            return self._const_vals[name];
        else:
            raise AttributeError('Unrecognized member of Immutable: %s' % name)

    def __init__(self, settable_vals, const_vals, lazy_vals, **opts):
        self.__dict__['_persistent'] = False
        self.__dict__['_settable_vals'] = make_dict(settable_vals).using(
            options=lambda s,o: make_dict(o))
        self.__dict__['_const_vals'] = make_dict(const_vals)
        self.__dict__['_lazy_vals'] = lazy_vals
        self.__dict__['_lazy_deps'] = self.__lazy_deps()
        self.options = opts
        return self

    def __repr__(self):
        return 'Immutable(<' + '>, <'.join([k for k,v in self._settable_vals.iteritems()]) + '>)'

    def transient(self):
        '''
        imm.transient() yields a transient copy of imm; i.e., the new immutable has all the same
        values as imm, but its 
        '''
        obj = copy.copy(self)
        obj.__dict__['_persistent'] = False
        return obj
    def persist(self):
        '''
        imm.persist() marks the immutable object imm as persistent then returns imm. Once an
        immutable is marked as persistent, its settable attributes may no longer be changed; to
        update values, instead create a new object with the transient() member function.
        '''
        self.__dict__['_persistent'] = True
        return self
    def is_persistent(self):
        '''
        imm.is_persistent() yields True if the immutable object imm is marked as persistent;
        otherwise yields False.
        '''
        return ('_persistent' in self.__dict__ and self.__dict__['_persistent'] is True)
    def using(self, **updates):
        '''
        imm.using(a=b...) yields a clone of the immutable object imm in which the values indicated 
        by the for the optional keywords (a) have been updated to the indicated values (b). The
        resulting immutable has a persistent state identical to imm, so if imm is transient, so will
        be the result.        
        '''
        obj = self.transient()
        for (k,v) in updates.iteritems():
            obj.__setattr__(k, v)
        if self._persistent:
            return obj.persist()
        else:
            return obj



####################################################################################################
# main/register_retinotopy.py
# The code for the function that handles the registration of retinotopy
# By Noah C. Benson

import numpy as np
import scipy as sp
import os, sys
from math import pi
from numbers import Number

import nibabel.freesurfer.io as fsio
import nibabel.freesurfer.mghformat as fsmgh
from pysistence import make_dict

from neuropythy.freesurfer import (freesurfer_subject, add_subject_path,
                                   cortex_to_ribbon, cortex_to_ribbon_map,
                                   Hemisphere)
from neuropythy.util import CommandLineParser
from neuropythy.vision import (register_retinotopy, retinotpy_model)


register_retinotopy_help = \
   '''
   The register_retinotopy command can be used to register a subject's
   hemisphere(s) to a model of V1-V3. At least one  subject id (either a freesurfer
   subject name, if SUBJECTS_DIR is set appropriately in the environment, or a path
   to a subject directory) must be given. Registration to a retinotopic model of
   V1-V3 is performed for both hemispheres of all of these subjects.
   In each subject's freesurfer directory, a variety of output data is deposited:
    * surf/lh.retinotopy_sym.sphere.reg
      xhemi/surf/lh.retinotopy_sym.sphere.reg
      These files contain the registrations of the left and right hemispheres to
      the retinotopy model. They are in the same format as Freesurfer's other 
      surface-data files such as surf/lh.white.
    * surf/lh.angle_predict.mgz   surf/rh.angle_predict.mgz
      surf/lh.eccen_predict.mgz   surf/rh.eccen_predict.mgz
      surf/lh.v123roi_predict.mgz surf/rh.v123roi_predict.mgz
      These files contain predictions of polar angle, eccentricity, and visual-area
      label for each hemisphere. The files are mgz format, so contain volumes;
      however, the volumes in each of these files is (1 x 1 x n) where n is the
      number of vertices in the hemisphere's Freesurfer meshes.
    * mri/angle_predict.mgz
      mri/eccen_predict.mgz
      mri/v123roi_predict.mgz
      These contain the data from the above surface data projected into the
      subject's 3D volume. Note that the volumes are oriented like Freesurfer's
      mri/brain.mgz file; if you want to convert this to the orientation of your
      original anatomical scan, use mri_convert:
       > mri_convert -rl mri/rawavg.mgz mri/angle_predict.mgz \\
                     mri/scanner.angle_predict.mgz
   The following options are accepted:
    * --eccen-lh=|-e<file>
      --angle-lh=|-a<file>
      --weight-lh=|-w<file>
      --eccen-rh=|-A<file>
      --angle-rh=|-E<file>
      --weight-rh=|-W<file>
      Each of these arguments specifies the name of a data file to load in as a
      representation of the subject's eccentricity, polar angle, or weight; these
      should be given the names of either an mgh/mgz files whose size is 1 x 1 x n,
      where n is the number of vertices in the hemisphere for the subject, or a
      FreeSurfer curv-style filename with n vertices. By default, files in the
      subject's surf directory that match a template are automatically loaded and
      used. This template is name <hemi>.<tag><name>, optionally ending with .mgz,
      where tag is one of (and in order of preference) 'prf_', 'empirical_',
      'measured_', 'training_', or '', and name is one of 'eccentricity'/'eccen',
      'polar_angle'/'angle', or 'weight'/'variance_explained'/'vexpl'.
    * --cutoff=|-c<value>
      The cutoff value to use for the weight; 0.1 by default. Weights less than
      this will be truncated to 0.
    * -N|--no-partial-correction
      Indicates that partial voluming correction should not be performed.
    * --angle-radians|-r
      This flag specifies that the angle-file only is in radians instead of
      degrees.
    * --eccen-radians|-R
      This flag specifies that the eccen-file only is in radians instead of
      degrees.
    * --mathematical|-m
      This flag specifies that the angle file addresses the visual space in the way
      standard in geometry; i.e., with the right horizontal meridian represented as
      0 and with the upper vertical meridian represented as 90 degrees or pi/4
      instead of the convention in which the opper vertical meridian represented as
      0 and the right horizontal meridian represented as 90 degrees or pi/4
      radians.
    * --edge-strength=|-D<weight>
      --angle-strength=|-T<weight>
      --functional-strength=|-F<weight>
      Each of these specifies the strength of the appropriate potential-field
      component. By default, these are each 1. Note that each field is already
      normalized by the number of components over which it operates; e.g., the edge
      strength is normalized by the number of edges in the mesh.
    * --max-steps=|-s<steps>
      This option specifies the maximum number of steps to run the registration; by
      default this is 2000.
    * --max-step-size=|-S<value>
      This specifies the max step-size for any single vertex; by default this is
      0.05.
    * --prior=|-p<name>
      This specifies the name of the prior registration to use in the fsaverage_sym
      subject; by default this is retinotopy. The prior may be omitted if the value
      "-" or "none" is given.
    * --eccen-tag=|-y<tag>
      --angle-tag=|-t<tag>
      --label-tag=|-l<tag>
      These options specify the output tag to use for the predicted measurement
      that results from the registration. By default, these are
      'eccen_predict', 'angle_predict', and 'v123roi_predict'.
      The output files have the name <hemi>.<tag>.mgz
    * --registration-name=|-u<string>
      This parameter indicates that the registration file, by default named 
      lh.retinotopy_sym.sphere.reg, should instead be named lh.<string>.sphere.reg.
    * --max-output-eccen=|-M<val>
      This specifies the maximum eccentricity to include in the output; there is no
      particular need to limit one's output, but it can be done with this argument.
      By default this is 90.
    * --no-volume-export|-x
      --no-surface-export|-z
      --no-registration-export|-X
      These flags indicate that the various data produced and written to the
      filesystem under normal execution should be suppressed. The volume export
      refers to the predicted volume files exported to the subject's mri directory;
      the registration export refers to the <hemi>.retinotopy_sym.sphere.reg file,
      written to the subject's surf directory, that contains the registered
      coordinates for the subject; and the surface export refers to the
      <hemi>.eccen_predict.mgz and similar files that are written to the
      subject's surf directory.
    * --subjects-dir=|-d
      Specifies additional subject directory search locations (in addition to the
      SUBJECTS_DIR environment variable and the FREESURFER_HOME/subjects
      directories, which are given here in descending search priority) when looking
      for subjects by name. This option cannot be specified multiple times, but it
      may contain : characters to separate directories, as in PATH.
    * --no-overwrite|-n
      This flag indicates that, when writing output files, no file should ever be
      replaced, should it already exist.
    * --
      This token, by itself, indicates that the arguments that remain should not be
      processed as flags or options, even if they begin with a -.
   '''
_retinotopy_parser_instructions = [
    # Flags
    ('h', 'help',                   'help',              False),
    ('v', 'verbose',                'verbose',           False),
    ('r', 'angle-radians',          'angle_radians',     False),
    ('R', 'eccen-radians',          'eccen_radians',     False),
    ('m', 'mathematical',           'angle_math',        False),
    ('x', 'no-volume-export',       'no_vol_export',     False),
    ('z', 'no-surface-export',      'no_surf_export',    False),
    ('X', 'no-registration-export', 'no_reg_export',     False),
    ('n', 'no-overwrite',           'no_overwrite',      False),
    ('N', 'no-partial-correction',  'part_vol_correct',  True),
    # Options                       
    ['e', 'eccen-lh',               'eccen_lh_file',     None],
    ['a', 'angle-lh',               'angle_lh_file',     None],
    ['w', 'weight-lh',              'weight_lh_file',    None],
    ['E', 'eccen-rh',               'eccen_rh_file',     None],
    ['A', 'angle-rh',               'angle_rh_file',     None],
    ['W', 'weight-rh',              'weight_rh_file',    None],
    ['c', 'cutoff',                 'weight_cutoff',     '0.1'],
    ['D', 'edge-strength',          'edge_strength',     '1'],
    ['T', 'angle-strength',         'angle_strength',    '1'],
    ['F', 'functional-strength',    'func_strength',     '1'],
    ['s', 'max-steps',              'max_steps',         '2000'],
    ['S', 'max-step-size',          'max_step_size',     '0.05'],
    ['p', 'prior',                  'prior',             'retinotopy'],
    ['y', 'eccen-tag',              'eccen_tag',         'eccen_predict'],
    ['t', 'angle-tag',              'angle_tag',         'angle_predict'],
    ['l', 'label-tag',              'label_tag',         'v123roi_predict'],
    ['u', 'registration-name',      'registration_name', 'retinotopy_sym'],
    ['M', 'max-output-eccen',       'max_out_eccen',     '90'],
    ['d', 'subjects-dir',           'subjects_dir',      None]]
_retinotopy_parser = CommandLineParser(_retinotopy_parser_instructions)
def _guess_surf_file(fl):
    if len(fl) > 4 and (fl[-4:] == '.mgz' or fl[-4:] == '.mgh'):
        return np.squeeze(np.array(fsmgh.load(fl).dataobj))
    else:
        return fsio.read_morph_data(fl)
def register_retinotopy_command(args):
    '''
    register_retinotopy_command(args) can be given a list of arguments, such as sys.argv[1:]; these
    arguments may include any options and must include at least one subject id. All subjects whose
    ids are given are registered to a retinotopy model, and the resulting registration, as well as
    the predictions made by the model in the registration, are exported.
    '''
    # Parse the arguments
    (args, opts) = _retinotopy_parser(args)
    # First, help?
    if opts['help']:
        print register_retinotopy_help
        return 1
    # and if we are verbose, lets setup a note function
    verbose = opts['verbose']
    def note(s):
        if verbose: print s
        return verbose
    # Add the subjects directory, if there is one
    if 'subjects_dir' in opts and opts['subjects_dir'] is not None:
        add_subject_path(opts['subjects_dir'])
    # Parse the simple numbers
    for o in ['weight_cutoff', 'edge_strength', 'angle_strength', 'func_strength',
              'max_step_size', 'max_out_eccen']:
        opts[o] = float(opts[o])
    opts['max_steps'] = int(opts['max_steps'])
    # These are for now not supported: #TODO
    if opts['angle_math'] or opts['angle_radians'] or opts['eccen_radians']:
        print 'Mathematical angles and angles not in degrees are not yet supported.'
        return 1
    # The remainder of the args can wait for now; walk through the subjects:
    tag_key = {'eccen': 'eccentricity', 'angle': 'polar_angle', 'label': 'V123_label'}
    for subnm in args:
        sub = freesurfer_subject(subnm)
        note('Processing subject: %s' % sub.id)
        # we need to register this subject...
        res = {}
        ow = not opts['no_overwrite']
        for h in ['LH','RH']:
            note('   Processing hemisphere: %s' % h)
            hemi = sub.__getattr__(h)
            # See if we are loading custom values...
            (ang,ecc,wgt) = (None,None,None)
            suffix = '_' + h.lower() + '_file'
            if opts['angle'  + suffix] is not None: ang = _guess_surf_file(opts['angle'  + suffix])
            if opts['eccen'  + suffix] is not None: ecc = _guess_surf_file(opts['eccen'  + suffix])
            if opts['weight' + suffix] is not None: wgt = _guess_surf_file(opts['weight' + suffix])
            # Do the registration
            note('    - Running Registration...')
            res[h] = register_retinotopy(hemi, retinotopy_model(),
                                         polar_angle=ang, eccentricity=ecc, weight=wgt,
                                         weight_cutoff=opts['weight_cutoff'],
                                         partial_voluming_correction=opts['part_vol_correct'],
                                         edge_scale=opts['edge_strength'],
                                         angle_scale=opts['angle_strength'],
                                         functional_scale=opts['func_strength'],
                                         prior=opts['prior'],
                                         max_predicted_eccen=opts['max_out_eccen'],
                                         max_steps=opts['max_steps'],
                                         max_step_size=opts['max_step_size'])
            # Perform the hemi-specific outputs now:
            if not opts['no_reg_export']:
                regnm = '.'.join([h.lower(), opts['registration_name'], 'sphere', 'reg'])
                flnm = (os.path.join(sub.directory, 'surf', regnm) if h == 'LH' else
                        os.path.join(sub.directory, 'xhemi', 'surf', regnm))
                if ow or not os.path.exist(flnm):
                    note('    - Exporting registration file: %s' % flnm)
                    fsio.write_geometry(flnm, res[h].coordinates.T, res[h].faces.T,
                                        'Created by neuropythy (github.com/noahbenson/neuropythy)')
                else:
                    note('    - Skipping registration file: %s (file exists)' % flnm)
            if not opts['no_surf_export']:
                for dim in ['angle', 'eccen', 'label']:
                    flnm = os.path.join(sub.directory, 'surf',
                                        '.'.join([h.lower(), opts[dim + '_tag'], 'mgz']))
                    if ow or not os.path.exist(flnm):
                        note('    - Exporting prediction file: %s' % flnm)
                        img = fsmgh.MGHImage(
                            np.asarray([[res[h].prop(tag_key[dim])]],
                                       dtype=(np.int32 if dim == 'label' else np.float32)),
                            np.eye(4))
                        img.to_filename(flnm)
                    else:
                        note('    - Skipping prediction file: %s (file exists)' % flnm)
        # Do the volume exports here
        if not opts['no_vol_export']:
            note('   Processing volume data...')
            note('    - Calculating cortex-to-ribbon mapping...')
            surf2rib = cortex_to_ribbon_map(sub, hemi=None)
            for dim in ['angle', 'eccen', 'label']:
                flnm = os.path.join(sub.directory, 'mri', opts[dim + '_tag'] + '.mgz')
                if ow or not os.path.exist(flnm):
                    note('    - Generating volume file: %s' % flnm)
                    vol = cortex_to_ribbon(sub,
                                           (res['LH'].prop(tag_key[dim]),
                                            res['RH'].prop(tag_key[dim])),
                                           map=surf2rib,
                                           dtype=(np.int32 if dim == 'label' else np.float32))
                    note('    - Exporting volume file: %s' % flnm)
                    vol.to_filename(flnm)
                else:
                    note('    - Skipping volume file: %s (file exists)' % flnm)
        # That is it for this subject!
        note('   Subject %s finished!' % sub.id)
    # And if we made it here, all was successful.
    return 0    




####################################################################################################
# neuropythy/util/command.py
# This file implements the command-line tools that are available as part of neuropythy.

from types      import (ListType, TupleType)
from pysistence import make_dict

class CommandLineParser(object):
    '''
    CommandLineParser(instructions) yields a command line parser object, which acts as a function,
    that is capable of parsing command line options into a dictionary of opts and a list of args,
    as defined by the list of instructions.
    The instructions should be a list of lists (or tuples), each of which should have three or four
    entries: [character, word, entry, default] where the default is optional.
    The character and word are the -x and --xxx versions of the argument; the entry is the name of
    the entry in the opts dictionary that is yielded for that command line option, and the default
    is the value inserted into the dictionary if the command line option is not found; if no
    default value is given, then the entry will appear in the dictionary only if it appears in the
    command line arguments.
    If the default value given is either True or False, then the option is understood to be a flag;
    i.e., the option does not take an argument (and single letter flags can appear together, such
    as -it instead of -i -t), and the appearance of the flag toggles the default to the opposite
    value.
    
    Example:
      parser = CommandLineParser(
        [('a', 'flag-a', 'aval', False),
         ('b', 'flag-b', 'bval', True),
         ('c', 'flag-c', 'cval', True),
         ('d', 'flag-d', 'dval', False),
         ('e', 'opt-e',  'eval', None),
         ('f', 'opt-f',  'fval', None),
         ('g', 'opt-g',  'gval'),
         ('h', 'opt-h',  'hval')])
      cmd_line = ['-ab', 'arg1', '--flag-d', '-etestE', '--opt-f=123', '-htestH', 'arg2']
      parser(cmd_line) == parser(*cmd_line)  # identical calls
      parser(cmd_line)
      # ==> (['arg1', 'arg2'],
      # ==>  {'a':True, 'b':False, 'c':True, 'd':True, 'e':'testE', 'f':'123', 'h':'testH'})
    '''

    def __init__(self, instructions):
        'See help(CommandLineParser).'
        wflags = {}
        cflags = {}
        wargs = {}
        cargs = {}
        defaults = {}
        for row in instructions:
            if not hasattr(row, '__iter__') or len(row) < 3 or len(row) > 4 or \
               any(x is not None and not isinstance(x, basestring) for x in row[:3]):
                raise ValueError('Invalid instruction row: %s ' % row)
            (c, w, var, dflt) = row if len(row) == 4 else (list(row) + [None])
            defaults[var] = dflt
            if dflt is True or dflt is False:
                if c is not None: cflags[c] = var
                if w is not None: wflags[w] = var
            else:
                if c is not None: cargs[c] = var
                if w is not None: wargs[w] = var
        self.default_values = make_dict(defaults)
        self.flag_words = make_dict(wflags)
        self.flag_characters = make_dict(cflags)
        self.option_words = make_dict(wargs)
        self.option_characters = make_dict(cargs)

    def __call__(self, *args):
        if len(args) > 0 and not isinstance(args[0], basestring) and \
           isinstance(args[0], (ListType, TupleType)):
            args = list(args)
            return self.__call__(*(list(args[0]) + args[1:]))
        parse_state = None
        more_opts = True
        remaining_args = []
        opts = dict(self.default_values)
        wflags = self.flag_words
        cflags = self.flag_characters
        wargs  = self.option_words
        cargs  = self.option_characters
        dflts  = self.default_values
        for arg in args:
            larg = arg.lower()
            if parse_state is not None:
                opts[parse_state] = arg
                parse_state = None
            else:
                if arg == '': pass
                elif more_opts and arg[0] == '-':
                    if len(arg) == 1:
                        remaining_args.append(arg)
                    elif arg[1] == '-':
                        trimmed = arg[2:]
                        if trimmed == '':     more_opts = False
                        if trimmed in wflags: opts[wflags[trimmed]] = not dflts[wflags[trimmed]]
                        else:
                            parts = trimmed.split('=')
                            if len(parts) == 1:
                                if trimmed not in wargs:
                                    raise ValueError('Unrecognized flag/option: %s' % trimmed)
                                # the next argument specifies this one
                                parse_state = wargs[trimmed]
                            else:
                                k = parts[0]
                                if k not in wargs:
                                    raise ValueError('Unrecognized option: %s' % k)
                                opts[wargs[k]] = trimmed[(len(k) + 1):]
                    else:
                        trimmed = arg[1:]
                        for (k,c) in enumerate(trimmed):
                            if c in cflags: opts[cflags[c]] = not dflts[cflags[c]]
                            elif c in cargs:
                                remainder = trimmed[(k+1):]
                                if len(remainder) > 0: opts[cargs[c]] = remainder
                                else:
                                    # next argument...
                                    parse_state = cargs[c]
                                break
                else:
                    remaining_args.append(arg)
        if parse_state is not None:
            raise ValueError('Ran out of arguments while awaiting value for %s' % parse_state)
        # that's done; all args are parsed
        return (remaining_args, opts)



#! /usr/bin/env python

import os
from setuptools import setup

# Deduce the version from the __init__.py file:
version = None
with open(os.path.join(os.path.dirname(__file__), 'neuropythy', '__init__.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break
if version is None: raise ValueError('No version found in neuropythy/__init__.py!')

setup(
    name='neuropythy',
    version=version,
    description='Toolbox for flexible cortical mesh analysis and registration',
    keywords='neuroscience mesh cortex registration',
    author='Noah C. Benson',
    author_email='nben@nyu.edu',
    maintainer_email='nben@nyu.edu',
    long_description='''
                     See the README.md file at the github repository for this package:
                     https://github.com/noahbenson/neuropythy
                     ''',
    url='https://github.com/noahbenson/neuropythy',
    download_url='https://github.com/noahbenson/neuropythy',
    license='GPLv3',
    classifiers=['Development Status :: 2 - Pre-Alpha',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: Software Development',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Information Analysis',
                 'Topic :: Scientific/Engineering :: Medical Science Apps.',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX',
                 'Operating System :: Unix',
                 'Operating System :: MacOS'],
    packages=['neuropythy',
              'neuropythy.util',
              'neuropythy.java',
              'neuropythy.geometry', 
              'neuropythy.immutable',
              'neuropythy.topology',
              'neuropythy.freesurfer',
              'neuropythy.cortex',
              'neuropythy.registration',
              'neuropythy.vision',
              'neuropythy.commands'],
    include_package_data=True,
    package_data={'': ['LICENSE.txt',
                       'neuropythy/lib/nben/target/nben-standalone.jar',
                       'neuropythy/lib/models/v123.fmm.gz',
                       'neuropythy/lib/models/benson17.fmm.gz',
                       'neuropythy/lib/data/fsaverage_sym/surf/lh.retinotopy_benson14.sphere.reg',
                       'neuropythy/lib/data/fsaverage_sym/surf/lh.retinotopy_benson17.sphere.reg',
                       'neuropythy/lib/data/fsaverage_sym/surf/sym.benson14_angle.mgz',
                       'neuropythy/lib/data/fsaverage_sym/surf/sym.benson14_eccen.mgz',
                       'neuropythy/lib/data/fsaverage_sym/surf/sym.benson14_varea.mgz',
                       'neuropythy/lib/data/fsaverage_sym/surf/sym.benson17_angle.mgz',
                       'neuropythy/lib/data/fsaverage_sym/surf/sym.benson17_eccen.mgz',
                       'neuropythy/lib/data/fsaverage_sym/surf/sym.benson17_varea.mgz']},
    install_requires=['numpy>=1.2',
                      'scipy>=0.7',
                      'nibabel>=2.0',
                      'pysistence>=0.4',
                      #'python-igraph>=0.7.1',
                      'py4j>=0.10'])



####################################################################################################
# neuropythy/freesurfer/subject.py
# Simple tools for use with FreeSurfer in Python
# By Noah C. Benson

import numpy as np
import numpy.linalg
import scipy as sp
import scipy.spatial as space
import nibabel.freesurfer.io as fsio
from   nibabel.freesurfer.mghformat import load as mghload, MGHImage
import os, math, copy, re
import itertools
from   warnings import warn
import pysistence
from   numbers import (Number, Integral)
from   types import TupleType
from   pysistence import make_dict

from   neuropythy.cortex    import (CorticalMesh)
from   neuropythy.topology  import (Topology, Registration)
from   neuropythy.immutable import (Immutable)
import neuropythy.geometry  as      geo

# These static functions are just handy
def spherical_distance(pt0, pt1):
    '''spherical_distance(a, b) yields the angular distance between points a and b, both of which
       should be expressed in spherical coordinates as (longitude, latitude).'''
    dtheta = pt1[0] - pt0[0]
    dphi   = pt1[1] - pt0[1]
    a = np.sin(dphi/2)**2 + np.cos(pt0[1]) * np.cos(pt1[1]) * np.sin(dtheta/2)**2
    return 2 * np.arcsin(np.sqrt(a))

class PropertyBox(list):
    '''PropertyBox is a simple class for lazily loading properties into a hemisphere or mesh. It
       should generally not be used directly and instead obtained from the Hemisphere class. Note
       that a PropertyBox object should behave like a list.'''
    def __init__(self, loader):
        self._loading_fn = loader
    def __len__(self):
        if 'data' not in self.__dict__: self.__load_data()
        return len(self.data)
    def __getitem__(self, key):
        if 'data' not in self.__dict__: self.__load_data()
        return self.data[key]
    def __iter__(self):
        if 'data' not in self.__dict__: self.__load_data()
        return iter(self.data)
    def __reversed__(self):
        if 'data' not in self.__dict__: self.__load_data()
        return reversed(self.data)
    def __contains__(self, item):
        if 'data' not in self.__dict__: self.__load_data()
        return item in self.data
    def __load_data(self):
        f = self._loading_fn
        dat = f()
        self.__dict__['data'] = dat

       
class Hemisphere(Immutable):
    '''
    The neuropythy.freesurfer.Hemisphere class inherits from neuropythy.Immutable and encapsulates
    the data contained in a subject's Freesurfer hemisphere. This includes the various surface data
    found in the Freesurfer subject's directory as well as certain volume data.
    '''
    

    ################################################################################################
    # Lazy/Static Interface
    # This code handles the lazy interface; only three data are non-lazy and these are coordinates,
    # faces, and options; they may be set directly.

    # The following methods and static variable allow one to set particular members of the object
    # directly:
    @staticmethod
    def _check_property(self, name, val):
        if not isinstance(val, np.ndarray) and not isinstance(val, list):
            raise ValueError('property values must be lists or numpy arrays')
        return True
    @staticmethod
    def _check_properties(self, val):
        if not isinstance(val, dict):
            raise ValueError('properties must be a dictionary')
        for (k, v) in val.iteritems(): Hemisphere._check_property(self, k, v)
        if isinstance(val, pysistence.persistent_dict.PDict):
            return val
        else:
            return make_dict(val)
    @staticmethod
    def _check_options(self, val):
        # Options just have to be a dictionary and are converted to an immutable one
        if not isinstance(val, dict):
            raise ValueError('options must be a dictionary')
        if isinstance(val, pysistence.persistent_dict.PDict):
            return val
        else:
            return make_dict(val)
    __settable_members = {
        'properties': lambda h,v: Hemisphere._check_properties(h,v),
        'options':    lambda h,v: Hemisphere._check_options(h,v)}
    
    # This static variable and these functions explain the dependency hierarchy in cached data
    def _make_surface(self, coords, faces, name, reg=None):
        mesh = CorticalMesh(
            coords,
            faces,
            subject = self.subject,
            hemisphere = self,
            meta_data = self.meta_data.using(
                **{'subject': self.subject,
                   'hemisphere': self,
                   'name': name,
                   'registration': reg}))
        if self.is_persistent(): mesh = mesh.persist()
        return mesh
    def _load_surface_data(self, name):
        path = name
        if not os.path.exists(path):
            path = self.subject.surface_path(name, self.name)
        if not os.path.exists(path):
            return None
        else:
            data = fsio.read_geometry(path)
            data[0].setflags(write=False)
            data[1].setflags(write=False)
            return data
    def _load_surface_data_safe(self, name):
        try:
            return self._load_surface_data(name)
        except:
            return (None, None)
    def _load_surface(self, name, preloaded=None, reg=None):
        data = self._load_surface_data(name) if preloaded is None else preloaded
        return self._make_surface(data[0], data[1], name, reg=reg)
    def _load_sym_surface(self, name):
        path = self.subject.surface_path(name, self.name)
        if not os.path.exists(path):
            return None
        else:
            data = fsio.read_geometry(path)
            data[0].setflags(write=False)
            data[1].setflags(write=False)
            return self._make_surface(data[0], data[1], name)
    def _load_ribbon(self):
        path = self.subject.volume_path('ribbon', self.name)
        if not os.path.exists(path):
            return None
        else:
            data = mghload(path)
            # for now, no post-processing, just loading of the MGHImage
            return data
    def _make_topology(self, faces, sphere, fsave, fssym):
        if sphere is None:
            return None
        if self.subject.id == 'fsaverage_sym':
            names = ['fsaverage_sym']
            surfs = [sphere]
        elif self.subject.id == 'fsaverage':
            names = ['fsaverage']
            surfs = [sphere]
        else:
            names = [self.subject.id, 'fsaverage', 'fsaverage_sym'];
            surfs = [sphere, fsave, fssym]
        regs = {names[i]: surfs[i] for i in range(len(names)) if surfs[i] is not None}
        # get the base path for the subject's surf directory and the prefix
        base = os.path.join(self.subject.directory, 'surf')
        prefix = self.chirality.lower() + '.'
        # find possible additional matches
        for fl in os.listdir(base):
            if len(fl) > 14 and fl.startswith(prefix) and fl.endswith('.sphere.reg'):
                data = self._load_surface_data_safe(fl[3:])
                if data[0] is not None:
                    regs[fl[3:-11]] = data[0]
        # If this is the fsaverage_sym and we didn't find the retinotopy topology, we need to
        # load it out of the data directory
        if self.subject.id == 'fsaverage_sym':
            path = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                'lib', 'data', 'fsaverage_sym', 'surf')
            if os.path.exists(path):
                regfls = {fl[3:-11]:os.path.join(path, fl) for fl in os.listdir(path)
                          if re.match('^lh\.[^\.]*\.sphere\.reg$', fl) is not None}
                for (regnm,regfl) in regfls.iteritems():
                    if regnm in regs: continue
                    data = self._load_surface_data_safe(regfl)
                    if data is None or data[0] is None:
                        warn('Could not load fsaverage_sym %s registration' % regnm)
                    else:
                        regs[regnm] = data[0]
        if self.subject.id == 'fsaverage_sym' and 'retinotopy' not in regs:
            if   'retinotopy_benson17' in regs: regs['retinotopy'] = regs['retinotopy_benson17']
            elif 'retinotopy_benson14' in regs: regs['retinotopy'] = regs['retinotopy_benson14']
            else: warn('Could not load fsaverage_sym retinotopy registration!')
        return Topology(faces, regs)
    @staticmethod
    def calculate_edge_data(faces):
        limit = max(faces.flatten()) + 1
        edge2face = {}
        idx = {}
        edge_list = [None for i in range(3*faces.size)]
        k = 0
        rng = range(faces.shape[1])
        for (e,i) in zip(
            zip(np.concatenate((faces[0], faces[1], faces[2])),
                np.concatenate((faces[1], faces[2], faces[0]))),
            np.concatenate((rng, rng, rng))):
            if e not in idx:
                idx[e] = k
                idx[e[::-1]] = k
                edge_list[k] = e
                edge2face[e] = i
                k += 1
            elif e in edge2face:
                edge2face[e[::-1]] = i
        for ((a,b),i) in edge2face.items():
            edge2face[(b,a)] = i
        return (np.array(edge_list[0:k]).transpose(), edge2face)
    @staticmethod
    def _check_meta_data(opts):
        md = opts.get('meta_data', {})
        if not isinstance(md, dict):
            raise ValueError('hemisphere meta-data must be a dictionary')
        if type(md) is pysistence.persistent_dict.PDict:
            return md
        else:
            return make_dict(**md)

    # This function will clear the lazily-evaluated members when a given value is changed
    def __update_values(self, name):
        for (sname, (deps, fn)) in Hemisphere.__lazy_members.items():
            if name in deps and sname in self.__dict__:
                del self.__dict__[sname]


        
    __lazy_members = {
        'meta_data':            (('options',), lambda opts: Hemisphere._check_meta_data(opts)),
        'property_names':       (('properties',), lambda props: set(props.keys())),
        'white_surface':        (('_load_surface',), lambda f: f('white')),
        'pial_surface':         (('_load_surface',), lambda f: f('pial')),
        'inflated_surface':     (('_load_surface',), lambda f: f('inflated')),

        'sphere_surface_data':  (('_load_surface_data',),      lambda f: f('sphere')),
        'fs_surface_data':      (('_load_surface_data_safe','subject'),
                                 lambda f,sub: (
                                     f('sphere') if sub.id == 'fsaverage' else 
                                     f('sphere.reg'))),
        'sym_surface_data':     (('_load_surface_data_safe','chirality','subject'), 
                                 lambda f, ch, sub: (
                                     f('sphere')                   if sub.id == 'fsaverage_sym' else
                                     f('fsaverage_sym.sphere.reg') if ch == 'LH'                else
                                     sub.RHX.sym_surface_data      if sub.RHX is not None       else
                                     (None,None))),
        'faces':                (('sphere_surface_data',), lambda dat: dat[1].T),
        'edge_data':            (('faces',), lambda F: Hemisphere.calculate_edge_data(F)),
        'edges':                (('edge_data',), lambda ED: ED[0]),
        'edge_face_index':      (('edge_data',), lambda ED: ED[1]),

        'sphere_surface':       (('_load_surface','subject','sphere_surface_data','topology'), 
                                 lambda f,sub,dat,topo: f(
                                     'sphere',
                                     preloaded=dat,
                                     reg=(topo.registrations[sub.id]
                                          if sub.id in topo.registrations
                                          else None))),
        'fs_sphere_surface':    (('_load_surface','fs_surface_data','topology'),
                                 lambda f,dat,topo: (
                                     None if dat == (None,None) else
                                     f('sphere.reg',
                                       preloaded=dat,
                                       reg=(topo.registrations['fsaverage']
                                            if 'fsaverage' in topo.registrations
                                            else None)))),
        'sym_sphere_surface':   (('_load_surface','sym_surface_data','topology'),
                                 lambda f,dat,topo: (
                                     None if dat is None else
                                     f('fsaverage_sym.sphere.reg',
                                       preloaded=dat,
                                       reg=(topo.registrations['fsaverage_sym']
                                            if 'fsaverage_sym' in topo.registrations
                                            else None)))),

        'vertex_count':         (('sphere_surface_data',), lambda dat: dat[0].shape[0]),
        'midgray_surface':      (('_make_surface', 'white_surface', 'pial_surface'),
                                 lambda f,W,P: f(
                                     0.5*(W.coordinates + P.coordinates),
                                     W.faces,
                                     'midgray')),
        'occipital_pole_index': (('inflated_surface',),
                                 lambda mesh: np.argmin(mesh.coordinates[1])),
        'ribbon':               (('_load_ribbon',), lambda f: f()),
        'chirality':            (('name',), 
                                 lambda name: 'LH' if name == 'LH' or name == 'RHX' else 'RH'),
        'topology':             (('_make_topology', 'sphere_surface_data',
                                  'fs_surface_data', 'sym_surface_data'),
                                 lambda f,sph,fs,sym: f(
                                     sph[1], sph[0],
                                     None if fs  is None or fs[0]  is None else fs[0],
                                     None if sym is None or sym[0] is None else sym[0]))}

    # Make a surface out of coordinates, if desired
    def surface(self, coords, name=None):
        '''
        hemi.surface(coords) yields a new surface with the given coordinate matrix and the topology
        of this hemisphere, hemi. The coordinate matrix must have hemi.vertex_count points.
        '''
        coords = np.asarray(coords)
        coords = (coords   if coords.shape[1] == self.vertex_count else
                  coords.T if coords.shape[0] == self.vertex_count else
                  None)
        if coords is None: raise ValueError('Coordinate matrix was invalid size!')
        return self._make_surface(coords, self.faces, name)
    def registration_mesh(self, name):
        '''
        hemi.registration_mesh(name) yields a CorticalMesh object for the given hemisphere hemi, as
        determined, by the name, which may be a registration name found in
        hemi.topology.registrations.
        Alternately, name may be a registration object, in which case it is used.
        '''
        if isinstance(name, Registration):
            return self.surface(name.coordinates)
        elif not isinstance(name, basestring) or name not in self.topology.registrations:
            raise ValueError('registration not found in topology')
        else:
            return self.surface(self.topology.registrations[name].coordinates, name=name)

    
    ################################################################################################
    # The Constructor
    def __init__(self, subject, hemi_name, **args):
        if not isinstance(subject, Subject):
            raise ValueError('Argument subject must be a FreeSurfer.Subject object')
        if hemi_name.upper() not in set(['RH', 'LH', 'RHX', 'LHX']):
            raise ValueError('Argument hemi_name must be RH, LH, RHX, or LHX')
        Immutable.__init__(self,
                           Hemisphere.__settable_members,
                           {'name':      hemi_name.upper(),
                            'subject':   subject,
                            'directory': os.path.join(subject.directory, 'surf')},
                           Hemisphere.__lazy_members)
        self.properties = args.pop('properties', make_dict())
        self.options = args
        self.__init_properties()
    
    ################################################################################################
    # The display function
    def __repr__(self):
        return "Hemisphere(" + self.name + ", <subject: " + self.subject.id + ">)"

    
    ################################################################################################
    # Property code
    def add_property(self, name, prop=Ellipsis):
        '''hemi.add_property(name, prop) adds (or overwrites) the given property with the given name
           in the given hemi. The name must be a valid dictionary key and the prop argument must be
           a list of numpy array of values, one per vertex.
           hemi.add_property(d) adds all of the properties in the given dictionary of properties, 
           d.
           Note that in either case, if the value of prop or the value in a dictionary item is
           None, then the item is removed from the property list instead of being added.'''
        if prop is Ellipsis:
            if isinstance(name, dict):
                for (n,p) in name.iteritems():
                    self.add_property(n, p)
            else:
                raise ValueError('add_property must be called with a name and propery or a dict')
        else:
            if prop is None:
                self.remove_property(name)
            else:
                self.properties = self.properties.using(**{name: prop})

    def has_property(self, name):
        '''hemi.has_property(name) yields True if the given hemisphere contains the property with
           the given name and False otherwise.'''
        return name in self.properties

    def remove_property(self, name):
        '''hemi.remove_property(name) removes the property with the given name from the given hemi.
           The name argument may also be an iterable collection of names, in which case all are
           removed.'''
        if hasattr(name, '__iter__'):
            for n in name: self.remove_property(n)
        elif name in self.properties:
            self.__dict__['properties'] = self.properties.without(name)
    
    def property_value(self, name):
        '''hemi.property_value(name) yields a list of the property values with the given name. If
           name is an iterable, then property_value automatically threads across it. If no such
           property is found in the mesh, then None is returned.'''
        if hasattr(name, '__iter__'):
            return map(lambda n: self.property_value(n), name)
        else:
            return self.properties.get(name, None)
            
    def prop(self, name, arg=Ellipsis):
        '''hemi.prop(...) is a generic function for handling mesh properties. It may be called
           with a variety of arguments:
             * hemi.prop(name) is equivalent to hemi.property_value(name) if name is either not
               an iterable or is an iterable that is not a dictionary
             * hemi.prop(d) is equivalent to hemi.add_property(d) if d is a dictionary
             * hemi.prop(name, val) is equivalent to hemi.add_property(name, val) if val is not
               None
             * hemi.prop(name, None) is equivalent to hemi.remove_property(name)'''
        if arg is Ellipsis:
            if isinstance(name, dict):
                self.add_property(name)
            else:
                return self.property_value(name)
        else:
            self.add_property(name, arg)

    def address(self, coords, registration=None, nearest=True):
        '''
        hemi.address(coords) yields the address dictionary of the given coords to their closest
        points in the 3D sphere registration of the given hemi. The optional argument registration
        may be specified to indicate that a different registration should be used; the default
        (None) indicates that the subject's native registration should be used. The optional
        argument nearest (default True) may also be set to False to indicate that the nearest point
        in the mesh should not be looked up
        '''
        coords = np.asarray(coords)
        if len(coords.shape) == 1:
            if coords.shape[0] == 0: return []
            else: return self.address([coords], registration=registration)[:,0]
        transposed = True if coords.shape[0] != 3 else False
        coords = coords.T if transposed else coords
        reg = self.topology.registrations[self.subject.id if registration is None else registration]
        if nearest: coords = reg.nearest(coords)
        return reg.address(coords)

    def unaddress(self, addrs, registration=None):
        '''
        hemi.unaddress(addrs) yields the coordinates of the given address dictionary addrs, looked
        up in the native registration of the given hemisphere hemi. The optional argument
        registration can be provided to indicate that a registration other than the subject's native
        registration should be used; the default (None) uses the native.
        '''
        reg = self.topology.registrations[self.subject.id if registration is None else registration]
        return reg.unaddress(addrs).T
            
    def interpolate(self, from_hemi, property_name, 
                    apply=True, method='automatic', mask=None, null=None, n_jobs=1):
        '''
        hemi.interpolate(from_hemi, prop) yields a list of property values that have been resampled
        onto the given hemisphere hemi from the property with the given name prop of the given 
        from_mesh. If the optional apply is set to True (the default) or to a new property name
        string, the property is added to hemi as well; otherwise it is only returned. Optionally,
        prop may be a list of numpy array of values the same length as the mesh's vertex list; in
        this case, the apply option must be a new property name string, otherwise it is treated as
        False. Note that in order to work, the hemi and from_hemi objects must share a registration
        such as fsaverage or fsaverage_sym.

        Options:
          * mask (default: None) indicates that the given True/False or 0/1 valued list/array should
            be used; any point whose nearest neighbor (see below) is in the given mask will, instead
            of an interpolated value, be set to the null value (see null option).
          * null (default: None) indicates the value that should be placed in the returned result if
            either a vertex does not lie in any triangle or a vertex is masked out via the mask
            option.
          * smoothing (default: 2) assuming that the method is 'interpolate' or 'automatic', this
            is the exponent used to smooth the interpolated surface; 1 is pure linear interpolation
            while 2 represents a slightly smoother version of this. Note that this is not an order
            of interpolation option.
          * method (default: 'automatic') specifies what method to use for interpolation. The only
            currently supported methods are 'automatic' or 'nearest'. The 'nearest' method does not
            actually perform a nearest-neighbor interpolation but rather assigns to a destination
            vertex the value of the source vertex whose veronoi-like polygon contains the
            destination vertex; note that the term 'veronoi-like' is used here because it uses the
            Voronoi diagram that corresponds to the triangle mesh and not the true delaunay
            triangulation. The 'automatic' checks every destination vertex and assigns it the
            'nearest' value if that value would not be a number, otherwise it interpolates linearly
            within the vertex's source triangle.
          * n_jobs (default: 1) is passed along to the cKDTree.query method, so may be set to an
            integer to specify how many processors to use, or may be -1 to specify all processors.
        '''
        if from_hemi.chirality != self.chirality:
            raise ValueError('hemispheres have opposite chiralities')
        if isinstance(property_name, basestring):
            if not from_hemi.has_property(property_name):
                raise ValueError('given property ' + property_name + ' is not in from_hemi!')
            data = from_hemi.prop(property_name)
        elif type(property_name).__module__ == np.__name__:
            data = property_name
            property_name = apply if isinstance(apply, basestring) else None
        elif hasattr(property_name, '__iter__'):
            if all(isinstance(s, basestring) for s in property_name):
                return {p: self.interpolate(from_hemi, p, 
                                            method=method, mask=mask,
                                            null=null, n_jobs=n_jobs)
                        for p in property_name}
            else:
                data = np.asarray(property_name)
                property_name = apply if isinstance(apply, basestring) else None
        else:
            raise ValueError('property_name is not a string or valid list')
        # pass data along to the topology object...
        result = self.topology.interpolate_from(from_hemi.topology, data,
                                                method=method, mask=mask,
                                                null=null, n_jobs=n_jobs)
        if result is not None:
            if apply is True and property_name is not None:
                self.prop(property_name, result)
            elif isinstance(apply, basestring):
                self.prop(apply, result)
        return result

    def partial_volume_factor(self, distance_cutoff=None, angle_cutoff=2.7):
        '''
        mesh.partial_volume_factor() yields an array of partial voluming risk metric values, one per
        vertex. Each value is a number between 0 and 1 such that a 1 indicates a very high risk for
        partial voluming and a 0 indicates a low risk.
        Partial voluming factors are calculated as follows:
        For each vertex u on the pial cortical surface, the set V of all vertices within d mm of u
        is found; d is chosen to be the mean edge length on the pial surface (if the option
        distance_cutoff is None) or the provided option.
        The partial volume factor for u, f(u) = length({v in V | (N(v) . N(u)) < k}) / length(V); the
        function N(u) indicates the vector normal to the cortical surface at vertex u and k is the
        cosine of the angle_cutoff option, which, by default, is 2.7 rad, or approximately 155 deg.
        '''
        pial = self.pial_surface
        normals = pial.vertex_normals.T
        d = np.mean(pial.edge_lengths) if distance_cutoff is None else distance_cutoff
        k = np.cos(angle_cutoff)
        # get the neighbors
        neis = pial.vertex_spatial_hash.query_ball_point(pial.coordinates.T, r=d)
        # calculate the fraction with large angles:
        return [float(len([1 for v in V if np.dot(u,normals[v]) < k])) / float(len(V))
                for (u,V) in zip(normals, neis)]
        
    
    # This [private] function and this variable set up automatic properties from the FS directory
    # in order to be auto-loaded, a property must appear in this dictionary:
    _auto_properties = {
        'sulc':             ('convexity',              lambda f: fsio.read_morph_data(f)),
        'thickness':        ('thickness',              lambda f: fsio.read_morph_data(f)),
        'area':             ('white_surface_area',     lambda f: fsio.read_morph_data(f)),
        'area.mid':         ('midgray_surface_area',   lambda f: fsio.read_morph_data(f)),
        'area.pial':        ('pial_surface_area',      lambda f: fsio.read_morph_data(f)),
        'curv':             ('curvature',              lambda f: fsio.read_morph_data(f)),

        'prf_eccen':        ('PRF_eccentricity',       lambda f: fsio.read_morph_data(f)),
        'prf_angle':        ('PRF_polar_angle',        lambda f: fsio.read_morph_data(f)),
        'prf_size':         ('PRF_size',               lambda f: fsio.read_morph_data(f)),
        'prf_varex':        ('PRF_variance_explained', lambda f: fsio.read_morph_data(f)),

        'retinotopy_eccen': ('eccentricity',           lambda f: fsio.read_morph_data(f)),
        'retinotopy_angle': ('polar_angle',            lambda f: fsio.read_morph_data(f)),
        'retinotopy_areas': ('visual_area',            lambda f: fsio.read_morph_data(f)),

        'retino_eccen':     ('eccentricity',           lambda f: fsio.read_morph_data(f)),
        'retino_angle':     ('polar_angle',            lambda f: fsio.read_morph_data(f)),
        'retino_areas':     ('visual_area',            lambda f: fsio.read_morph_data(f))}

    # properties grabbed out of MGH or MGZ files
    _mgh_properties = {
        'prf_eccen':         ('PRF_eccentricity',       lambda f: mghload(f).get_data().flatten()),
        'prf_angle':         ('PRF_polar_angle',        lambda f: mghload(f).get_data().flatten()),
        'prf_size':          ('PRF_size',               lambda f: mghload(f).get_data().flatten()),
        'prf_varex':         ('PRF_variance_explained', lambda f: mghload(f).get_data().flatten()),
        'prf_vexpl':         ('PRF_variance_explained', lambda f: mghload(f).get_data().flatten()),

        'retinotopy_eccen':  ('eccentricity',           lambda f: fsio.read_morph_data(f)),
        'retinotopy_angle':  ('polar_angle',            lambda f: fsio.read_morph_data(f)),
        'retinotopy_areas':  ('visual_area',            lambda f: fsio.read_morph_data(f)),
        'retino_eccen':      ('eccentricity',           lambda f: fsio.read_morph_data(f)),
        'retino_angle':      ('polar_angle',            lambda f: fsio.read_morph_data(f)),
        'retino_areas':      ('visual_area',            lambda f: fsio.read_morph_data(f)),

        'predicted_eccen':   ('predicted_eccentricity', lambda f: mghload(f).get_data().flatten()),
        'predicted_angle':   ('predicted_polar_angle',  lambda f: mghload(f).get_data().flatten()),
        'predicted_areas':   ('predicted_visual_area',  lambda f: mghload(f).get_data().flatten()),
        'predicted_v123roi': ('predicted_visual_area',  lambda f: mghload(f).get_data().flatten()),

        'predict_eccen':     ('predicted_eccentricity', lambda f: mghload(f).get_data().flatten()),
        'predict_angle':     ('predicted_polar_angle',  lambda f: mghload(f).get_data().flatten()),
        'predict_areas':     ('predicted_visual_area',  lambda f: mghload(f).get_data().flatten()),
        'predict_v123roi':   ('predicted_visual_area',  lambda f: mghload(f).get_data().flatten()),

        'benson14_eccen':    ('benson14_eccentricity',  lambda f: mghload(f).get_data().flatten()),
        'benson14_angle':    ('benson14_polar_angle',   lambda f: mghload(f).get_data().flatten()),
        'benson14_areas':    ('benson14_visual_area',   lambda f: mghload(f).get_data().flatten()),
        'benson14_v123roi':  ('benson14_visual_area',   lambda f: mghload(f).get_data().flatten()),

        'eccen_predict':     ('predicted_eccentricity', lambda f: mghload(f).get_data().flatten()),
        'angle_predict':     ('predicted_polar_angle',  lambda f: mghload(f).get_data().flatten()),
        'areas_predict':     ('predicted_visual_area',  lambda f: mghload(f).get_data().flatten()),
        'v123roi_predict':   ('predicted_visual_area',  lambda f: mghload(f).get_data().flatten()),

        'eccen_benson14':    ('benson14_eccentricity',  lambda f: mghload(f).get_data().flatten()),
        'angle_benson14':    ('benson14_polar_angle',   lambda f: mghload(f).get_data().flatten()),
        'areas_benson14':    ('benson14_visual_area',   lambda f: mghload(f).get_data().flatten()),
        'v123roi_benson14':  ('benson14_visual_area',   lambda f: mghload(f).get_data().flatten())}
    
    # funciton for initializing the auto-loading properties
    def __init_properties(self):
        # if this is an xhemi, we want to load the opposite of the chirality
        loadchi = 'RH' if self.name == 'RHX' else \
                  'LH' if self.name == 'LHX' else \
                  self.chirality
        dir = self.directory
        files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
        autoprops = Hemisphere._auto_properties
        for file in files:
            if len(file) > 2 and file[0:2].upper() == loadchi and file[3:] in autoprops:
                (name, fn) = autoprops[file[3:]]
                #self.prop(name, PropertyBox(lambda: fn(os.path.join(dir, file))))
                self.prop(name, fn(os.path.join(dir, file)))
        # We also want to auto-add labels:
        dir = os.path.join(self.subject.directory, 'label')
        files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
        for file in files:
            if len(file) > 9 and file[0:2].upper() == loadchi and file[-6:] == '.label':
                if len(file) < 17 or file[-13:-6] != '.thresh':
                    lbl = set(fsio.read_label(os.path.join(dir, file)))
                    self.prop(
                        file[3:-6],
                        [True if k in lbl else False for k in range(self.vertex_count)])
                #else:
                #    (lbl, sclr) = fsio.read_label(os.path.join(dir, file), read_scalars=True)
                #    lbl = {lbl[i]: i for i in range(len(lbl))}
                #    self.prop(
                #        file[3:-13] + '_threshold',
                #        [sclr[lbl[k]] if k in lbl else None for k in range(self.vertex_count)])
        # And MGH/MGZ files in surf:
        dir = os.path.join(self.subject.directory, 'surf')
        files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
        for f in files:
            if (f[-4:].lower() == '.mgh' or f[-4:].lower() == '.mgz') and \
                    f[2] == '.' and f[0:2].upper() == loadchi and \
                    f[3:-4] in Hemisphere._mgh_properties:
                (name, fn) = Hemisphere._mgh_properties[f[3:-4]]
                try:
                    pvals = fn(os.path.join(dir, f))
                except:
                    pvals = None
                if pvals is not None:
                    self.prop(name, pvals)

    # This method is a convenient way to get the occipital pole coordinates for the various
    # surfaces in a hemisphere...
    def occipital_pole(self, surfname):
        surfname = surfname.lower()
        if not surfname.lower().endswith('_surface'):
            surfname = surfname + '_surface'
        surf = self.__getattr__(surfname)
        opIdx = self.occipital_pole_index
        return surf.coordinates[:, opIdx]

    # These methods, values, and the projection fn are all related to map projection
    @staticmethod
    def _orthographic_projection(X, params): 
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        return X[1:3]
    @staticmethod
    def _orthographic_projection_inverse(X, params): 
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        Xnorm = X / r
        return np.asarray([r * np.sqrt(1.0 - (Xnorm ** 2).sum(0)), X[0], X[1]])
    @staticmethod
    def _equirectangular_projection(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        X = X / np.sqrt((X ** 2).sum(0))
        r = params['sphere_radius']
        return r / math.pi * np.asarray([np.arctan2(X[1], X[0]), np.arcsin(X[2])])
    @staticmethod
    def _equirectangular_projection_inverse(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        X = math.pi / r * X
        cos1 = np.cos(X[1])
        return np.asarray([cos1 * np.cos(X[0]) * r, 
                           cos1 * np.sin(X[0]) * r,
                           np.sin(X[1]) * r])
    @staticmethod
    def _mercator_projection(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        X = X / np.sqrt((X ** 2).sum(0))
        r = params['sphere_radius']
        return r * np.asarray([np.arctan2(X[1], X[0]),
                               np.log(np.tan(0.25 * math.pi + 0.5 * np.arcsin(X[2])))])
    @staticmethod
    def _mercator_projection_inverse(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        X = X / r
        return r * np.asarray([np.cos(X[0]),
                               np.sin(X[0]),
                               np.sin(2 * (np.arctan(np.exp(X[1])) - 0.25*math.pi))])
    @staticmethod
    def _sinusoidal_projection(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        X = X / np.sqrt((X ** 2).sum(0))
        r = params['sphere_radius']
        phi = np.arcsin(X[2])
        return r / math.pi * np.asarray([np.arctan2(X[1], X[0]) * np.cos(phi), phi])
    @staticmethod
    def _sinusoidal_projection_inverse(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        X = math.pi * X / r
        z = np.sin(X[1])
        cosphi = np.cos(X[1])
        return np.asarray([np.cos(X[0] / cosphi) * r,
                           np.sin(X[0] / cosphi) * r,
                           np.sin(X[1]) * r])

    __projection_methods = {
        'orthographic':    lambda X,p: Hemisphere._orthographic_projection(X,p),
        'equirectangular': lambda X,p: Hemisphere._equirectangular_projection(X,p),
        #'mollweide':       lambda X,p: Hemisphere._mollweide_projection(X,p),
        'mercator':        lambda X,p: Hemisphere._mercator_projection(X,p),
        'sinusoidal':      lambda X,p: Hemisphere._sinusoidal_projection(X,p)}
    __projection_methods_inverse = {
        'orthographic':    lambda X,p: Hemisphere._orthographic_projection_inverse(X,p),
        'equirectangular': lambda X,p: Hemisphere._equirectangular_projection_inverse(X,p),
        #'mollweide':       lambda X,p: None,
        'mercator':        lambda X,p: Hemisphere._mercator_projection_inverse(X,p),
        'sinusoidal':      lambda X,p: Hemisphere._sinusoidal_projection_inverse(X,p)}

    @staticmethod
    def _make_projection(params):
        '''
        Create the projection forward and inverse functions, based on the parameters, which should
        be identical to those given in the Hemisphere.projection() and Hemisphere.projection_data()
        funtcions, except that automatic values such as None that get interpreted by the hemisphere
        object should all already be filled out explicitly. Yields a persistent dictionary that is
        identical to the given params argument map, but includes both an inverse and a forward
        function in the entries 'inverse_function' and 'forward_function'.
        '''
        registration = 'native' if 'registration' not in params else params['registration']
        if not isinstance(registration, basestring):
            raise ValueError('registration parameter must be a string')
        center = None if 'center' not in params else params['center']
        if not hasattr(center, '__iter__') or len(center) < 2 or len(center) > 3:
            raise ValueError('center argument must be explicit and must be a 2 or 3-length vector')
        center = np.asarray(center)
        if len(center) == 2:
            # convert to 3D cartesian coordinates from the, presumed, longitude and latitude in
            # angular coordinates (radians)
            center_sc = center
            cos_phi = np.cos(center[1])
            center = np.asarray([cos_phi * np.cos(center[0]),
                                 cos_phi * np.sin(center[0]),
                                 np.sin(center[1])])
        else:
          center_sc = [np.arctan2(center[0], center[1]),
                       np.arcsin(center[2] / np.sqrt(sum(center ** 2)))]
        center_right = [0,1,0] if 'center_right' not in params else params['center_right']
        if not hasattr(center_right, '__iter__') or len(center_right) < 2 or len(center_right) > 3:
            raise ValueError('center_right argument must be explicit and must be a 2 or 3-length' \
                             + ' vector')
        center_right = np.asarray(center_right)
        if len(center_right) == 2:
            center_right_sc = center_right
            cos_phi = np.cos(center_right[1])
            center_right = np.asarray([cos_phi * np.cos(center_right[0]),
                                       cos_phi * np.sin(center_right[0]),
                                       np.sin(center[1])])
        radius = None if 'radius' not in params else params['radius']
        if not (isinstance(radius, Number) or np.issubdtype(type(radius), np.float)):
            raise ValueError('radius option must be explicitly given and must be a number')
        radius = abs(radius)
        sphere_radius = None if 'sphere_radius' not in params else params['sphere_radius']
        if not (isinstance(sphere_radius, Number) or np.issubdtype(type(sphere_radius), np.float)):
            raise ValueError('sphere_radius option must be explicitly given and must be a number')
        sphere_radius = abs(sphere_radius)
        method = None if 'method' not in params else params['method']
        if not isinstance(method, basestring) or method not in Hemisphere.__projection_methods:
            raise ValueError('method parameter not given or not recognized')
        method = method.lower()
        chirality = None if 'chirality' not in params else params['chirality']
        if chirality is not None:
            if not isinstance(chirality, basestring):
                raise ValueError('chirality must be None or a string representing left or right')
            chirality = chirality.upper()
            if chirality == 'L' or chirality == 'R':
                chirality += 'H'
            elif chirality == 'LEFT' or chirality == 'RIGHT':
                chirality = chirality[0] + 'H'
            elif chirality != 'LH' and chirality != 'RH':
                raise ValueError('Chiraliry must be one of LH, RH, L, R, Left, Right, or None')
        params['center'] = center
        params['center_right'] = center_right
        params['center_spherical'] = center_sc
        params['radius'] = radius
        params['sphere_radius'] = sphere_radius
        params['method'] = method
        params['chirality'] = chirality
        params['registration'] = registration
        # Setup Projection Data:
        FR = geo.alignment_matrix_3D(center, [1.0, 0.0, 0.0])
        cr = np.dot(FR, center_right)
        rot_ang = math.pi/2 -  np.arctan2(cr[1], -cr[2])
        FR = np.dot(geo.rotation_matrix_3D([1.0,0.0,0.0], rot_ang), FR)
        IR = numpy.linalg.inv(FR)
        params['forward_affine_transform'] = FR
        params['inverse_affine_transform'] = IR
        fwdprojfn = Hemisphere.__projection_methods[method]
        invprojfn = None if method not in Hemisphere.__projection_methods_inverse else \
                    Hemisphere.__projection_methods_inverse[method]
        def __fwdfn(obj):
            # in this forward-function, we want to interpret what the argument actually is...
            if isinstance(obj, Hemisphere):
                # check to make sure it's the right chirality
                if chirality is not None and obj.chirality != chirality:
                    raise ValueError('Cannot project hemisphere if opposite chirality')
                # See if it has the appropriate registration
                usereg = obj.subject.id if registration == 'native' else registration
                if usereg not in obj.topology.registrations:
                    raise ValueError('Given hemisphere is not registered to the ' + registration \
                                     + ' registration')
                reg = obj.topology.registrations[usereg]
                mesh = obj._make_surface(reg.coordinates, obj.topology.triangles, usereg)
                proj = __fwdfn(mesh)
                proj_params = proj.options['projection_parameters']
                proj.options = proj.options.using(
                    projection_parameters=proj_params.using(hemisphere=obj))
                return proj
            elif isinstance(obj, CorticalMesh):
                sc = obj.spherical_coordinates
                submesh = obj.select(lambda u: spherical_distance(center_sc, sc[0:2,u]) < radius)
                submesh.coordinates = __fwdfn(submesh.coordinates)
                submesh.options = submesh.options.using(
                    projection_parameters=params.using(mesh=obj))
                return submesh
            else:
                obj = np.asarray(obj)
                X = obj if obj.shape[0] == 3 else obj.T
                return fwdprojfn(np.dot(FR, X), params)
        def __invfn(obj):
            if isinstance(obj, CorticalMesh):
                obj_proj_params = obj.options['projection_parameters']
                mesh = obj_proj_params['mesh']
                X = mesh.coordinates.copy()
                X[:, obj.vertex_labels] = __invfn(obj.coordinates)
                X.flags.writeable = False
                return mesh.using(coordinates=X)
            else:
                X = np.asarray(obj)
                X = X if X.shape[0] == 2 else X.T
                return np.dot(IR, invprojfn(X, params))
        params['forward_function'] = __fwdfn
        params['inverse_function'] = __invfn
        params = make_dict(params)
        return params
            
    def __interpret_projection_params(self, params):
        params = params.copy()
        # Figure out which spherical surface we're using...
        regname = None if 'registration' not in params else params['registration'].lower()
        if regname is None or regname == 'native':
            regname = self.subject.id
        elif regname not in self.topology.registrations:
            raise ValueError('Unrecognized registration name: %s' % regname)
        params['registration'] = regname
        coords = self.topology.registrations[regname].coordinates.T
        params['sphere_radius'] = np.mean(np.sqrt(np.sum(coords**2, 0)))
        # important parameters: center and radius => indices
        if 'center' not in params:
            params['center'] = coords[:, self.occipital_pole_index]
        elif isinstance(params['center'], int):
            params['center'] = coords[:, params['center']]
        if 'radius' not in params:
            params['radius'] = math.pi / 3.0
        # also, we need to worry about the method...
        if 'method' not in params:
            params['method'] = 'equirectangular'
        # And the chirality...
        if 'chirality' not in params:
            params['chirality'] = self.chirality
        return params

    def projection_data(self, **params):
        '''
        hemi.projection_data() yields an immutable dictionary of data describing a map projection
        of the given hemisphere hemi.

        The following options are understood; options that are not understood are stored as 
        meta-data and placed in the projection_parameters entry of resulting projected meshes' 
        options member variable:
          * center (default: use the occipital pole) specifies the 3D cartesian or 2D spherical 
            (longitude, latitude, both in radians) vector that is the center of the map projection.
          * radius (default: pi/3) specifies the distance (in radians along the spherical surface)
            within which vertices should be included in the projection.
          * method (default: 'equirectangular') specifies the map projection type; see also
            Hemisphere._make_projection.
          * registration (default: self.subject.id) specifies the registration to which the map
            should be aligned when reprojecting.
          * chirality (default: self.chirality) specifies which chirality the projection should be
            restricted to when reprojecting; if None, then chirality is not restricted.
        '''
        params = self.__interpret_projection_params(params)
        return Hemisphere._make_projection(params)
    
    def projection(self, **params):
        '''
        hemi.projection() yields a map projection that is centered at the occipital pole of the
        hemisphere's spherical topology registration (also hemi.sphere_surface). A variety of
        options that modify the projection may be given; these options are detailed in the
        Hemisphere.projection_data method.
        '''
        proj = self.projection_data(**params)
        return proj['forward_function'](self)

# Setup the subject directory:
_subjects_dirs = []
def subject_paths():
    '''
    subject_paths() yields a list of paths to Freesurfer subject directories in which subjects are
    automatically searched for when identified by subject-name only. These paths are searched in
    the order returned from this function.
    If you must edit these paths, it is recommended to use add_subject_path, but the _subjects_dirs
    variable may be set as well.
    '''
    return copy.copy(_subjects_dirs)

def add_subject_path(path, index=0):
    '''
    add_subject_path(path) will add the given path to the list of subject directories in which to
    search for Freesurfer subjects. The optional argument index may be given to specify the
    precedence of this path when searching for new subject; the default, 0, always inserts the path
    at the front of the list; a value of k indicates that the new list should have the new path at
    index k.
    The path may contain :'s, in which case the individual directories are separated and added.
    If the given path is not a directory or the path could not be inserted, yields False;
    otherwise, yields True. If the string contains a : and multiple paths, then True is yielded
    only if all paths were successfully inserted.
    See also subject_paths.
    '''
    paths = [p for p in path.split(':') if len(p) > 0]
    if len(paths) > 1:
        tests = [add_subject_path(p, index=index) for p in reversed(paths)]
        return all(t for t in tests)
    else:
        if not os.path.isdir(path): return False
        if path in _subjects_dirs:   return True
        try:
            _subjects_dirs.insert(index, path)
            return True
        except:
            return False
# add the SUBJECTS_DIR environment variable...
if 'FREESURFER_HOME' in os.environ:
    add_subject_path(os.path.join(os.environ['FREESURFER_HOME'], 'subjects'))
if 'SUBJECTS_DIR' in os.environ:
    add_subject_path(os.environ['SUBJECTS_DIR'])
    
def find_subject_path(sub):
    '''
    find_subject_path(sub) yields the full path of a Freesurfer subject with the name given by the
    string sub, if such a subject can be found in the Freesurfer search paths. See also
    add_subject_path.
    If no subject is found, then None is returned.
    '''
    looks_like_sub = lambda d: all(os.path.isdir(os.path.join(d, sfx) if sfx is not None else d)
                                   for sfx in [None, 'mri', 'surf', 'label'])
    # if it's a full/relative path already, use it:
    if looks_like_sub(sub): return sub
    for sdir in _subjects_dirs:
        tmp = os.path.join(sdir, sub)
        if looks_like_sub(tmp): return tmp
    return None
    
class Subject(Immutable):
    '''FreeSurfer.Subject objects encapsulate the data contained in a FreeSurfer
       subject directory.'''

    ################################################################################################
    # Lazy/Static Interface
    # This code handles the lazy interface; only three data are non-lazy and these are coordinates,
    # faces, and options; they may be set directly.

    # The following methods and static variable allow one to set particular members of the object
    # directly:
    @staticmethod
    def _check_options(self, val):
        # Options just have to be a dictionary and are converted to an immutable one
        if not isinstance(val, dict):
            raise ValueError('options must be a dictionary')
        if type(val) is pysistence.persistent_dict.PDict:
            return val
        else:
            return make_dict(**val)
    __settable_members = {
        'options': lambda m,v: Subject._check_options(m,v)}

    # This static variable and these functions explain the dependency hierarchy in cached data
    @staticmethod
    def _check_meta_data(opts):
        md = opts.get('meta_data', {})
        if not isinstance(md, dict):
            raise ValueError('subject meta-data must be a dictionary')
        if type(md) is pysistence.persistent_dict.PDict:
            return md
        else:
            return make_dict(**md)
    def _load_hemisphere(self, name):
        hem = Hemisphere(self, name)
        if self.is_persistent(): hem.persist()
        return hem
    def _load_hemisphere_safe(self, name):
        try:
            hem = Hemisphere(self, name)
            if self.is_persistent(): hem.persist()
            return hem
        except:
            return None
    __lazy_members = {
        'meta_data': (('options',), lambda opts: Subject._check_meta_data(opts)),
        'LH':  (('_load_hemisphere',),      lambda f: f('LH')),
        'RH':  (('_load_hemisphere',),      lambda f: f('RH')),
        'LHX': (('_load_hemisphere_safe',), lambda f: f('LHX')),
        'RHX': (('_load_hemisphere_safe',), lambda f: f('RHX'))}
    
    
    ################################################################################################
    # The Constructor
    def __init__(self, subject, **args):
        subpath = find_subject_path(subject)
        if subpath is None: raise ValueError('No valid subject found: %s' % subject)
        (dr, name) = os.path.split(subpath)
        Immutable.__init__(self,
                           Subject.__settable_members,
                           {'subjects_dir': dr,
                            'id': name,
                            'directory': subpath},
                           Subject.__lazy_members)
        self.options = args
    
    ################################################################################################
    # The display function
    def __repr__(self):
        return "Subject(" + self.id + ")"


    ################################################################################################
    # Standard Methods

    def surface_path(self, type, hemi):
        hemi = hemi.upper()
        if hemi == 'RHX':
            if not os.path.isdir(os.path.join(self.directory, 'xhemi', 'surf')):
                raise ValueError('Subject does not have an xhemi/surf directory')
            path = os.path.join(self.directory, 'xhemi', 'surf', 'lh.' + type)
        elif hemi == 'LHX':
            if not os.path.isdir(os.path.join(self.directory, 'xhemi', 'surf')):
                raise ValueError('Subject does not have an xhemi/surf directory')
            path = os.path.join(self.directory, 'xhemi', 'surf', 'rh.' + type)
        elif hemi == 'LH':
            path = os.path.join(self.directory, 'surf', 'lh.' + type)
        elif hemi == 'RH':
            path = os.path.join(self.directory, 'surf', 'rh.' + type)
        else:
            raise ValueError('Argument hemi must be LH, RH, LHX, or RHX')
        if not os.path.isfile(path):
            raise ValueError('file %s not found' % path)
        return path
    def volume_path(self, type, hemi):
        # hemi may be None, in which case we have only these surfaces:
        if hemi is None:
            nm = type + '.'
        else:
            nm = hemi.lower() + '.' + type + '.'
        flnm = os.path.join(self.directory, 'mri', nm + 'mgz')
        if not os.path.isfile(flnm):
            flnm = os.path.join(self.directory, 'mri', nm + 'mgh')
            if not os.path.isfile(flnm):
                raise ValueError('no such file (or mgz equivalent) found: ' + flnm)
        return flnm
            

####################################################################################################
# Some FreeSurfer specific functions

def cortex_to_ribbon_map_smooth(sub, hemi=None, k=12, distance=4, sigma=0.35355):
    '''cortex_to_ribbon_map_smooth(sub) yields a dictionary whose keys are the indices of the ribbon
         voxels for the given FreeSurfer subject sub and whose values are a tuple of both (0) a list
         of vertex labels associated with the given voxel and (1) a list of the weights associated
         with each vertex (both in identical order).

       These associations are determined by finding the k nearest neighbors (of the midgray surface)
       to each voxel then finding the lines from white-to-pial that pass within a certain distance
       of the voxel center and weighting each appropriately.

       The following options are accepted:
         * k (default: 4) is the number of vertices to find near each voxel center
         * distance (default: 3) is the cutoff for including a vertex in a voxel region
         * hemi (default: None) specifies which hemisphere to operate over; this may be 'lh', 'rh',
           or None (to do both)
         * sigma (default: 1 / (2 sqrt(2))) specifies the standard deviation of the Gaussian used
           to weight the vertices contributing to a voxel'''
    
    # we can speed things up slightly by doing left and right hemispheres separately
    if hemi is None:
        return (cortex_to_ribbon_map(sub, k=k, distance=distance, sigma=sigma, hemi='lh'),
                cortex_to_ribbon_map(sub, k=k, distance=distance, sigma=sigma, hemi='rh'))
    if not isinstance(hemi, basestring):
        raise ValueError('hemi must be a string \'lh\', \'rh\', or None')
    if hemi.lower() == 'lh' or hemi.lower() == 'left':
        hemi = sub.LH
    elif hemi.lower() == 'rh' or hemi.lower() == 'right':
        hemi = sub.RH

    # first we find the indices of all voxels that are in the ribbon
    ribdat = hemi.ribbon.get_data()
    idcs = np.transpose(ribdat.nonzero()) + 1  # voxel transforms assume 1-based indexing
    # we also need the transformation from surface to voxel
    #tmtx = hemi.ribbon.header.get_ras2vox()
    tmtx = np.linalg.inv(hemi.ribbon.header.get_vox2ras_tkr())

    # now we want to make an octree from the midgray voxels
    txcoord = lambda mtx: np.dot(tmtx[0:3], np.vstack((mtx, np.ones(mtx.shape[1]))))
    midgray = txcoord(hemi.midgray_surface.coordinates)
    near = space.KDTree(midgray.T)
    
    # then we find the k nearest vertices for each voxel with a distance cutoff
    (d, nei) = near.query(idcs, k=k, distance_upper_bound=distance, p=2)
    ## we want the i'th row of d and nei to be a list of the i'th-nearests to each voxel center
    d = d.T
    nei = nei.T
    
    # okay, now we want to find the closest point along each line segment... we do this iteratively;
    # in the process we replace the d array with the distance of the nearest point along the segment
    # between pial and white for the given vertex. Any inf value will be left alone and will become
    # a zero when run through exp

    ## grab the pial and white matter coordinates...
    pialX = txcoord(hemi.pial_surface.coordinates)
    whiteX = txcoord(hemi.white_surface.coordinates)
    pial2white = whiteX - pialX
    pial2white_norm = np.sqrt(np.sum(pial2white ** 2, 0))
    same_idcs = np.where(pial2white_norm < 1e-9)
    pial2white_norm[same_idcs] = 1
    pial2white[:, same_idcs] = 0
    pial2white_normed = pial2white / pial2white_norm
    ## At this point we also want a transpose of indices
    idcs = idcs.T
    ## Iterate over the neighbors
    for i in range(d.shape[0]):
        # which elements are finite? (which do we care about)
        (finite_vox_idcs,) = np.where(d[i] != np.inf)
        # we do a closest-point search along the segment from white to pial in these locations
        ## to start, we project out the normal
        finite_vtx_idcs = nei[i, finite_vox_idcs]
        finite_vox_inplane = idcs[:, finite_vox_idcs] - pialX[:, finite_vtx_idcs]
        finite_p2w_normed = pial2white_normed[:, finite_vtx_idcs]
        finite_vox_prods = np.sum(finite_vox_inplane * finite_p2w_normed, 0)
        finite_vox_ipnorm = np.sqrt(np.sum(finite_vox_inplane ** 2, 0))
        ## If the product is less than 0, that means the voxel center is outside of the pial surface
        (inval,) = np.where((finite_vox_prods < -0.1 * finite_vox_ipnorm) \
                                + (finite_vox_prods > 1.1 * finite_vox_ipnorm))
        if len(inval) > 0:
            # annotate that these are invalid globally
            d[i, finite_vox_idcs[inval]] = np.inf
            # and locally
            finite_vox_idcs = np.delete(finite_vox_idcs, inval)
            finite_vox_inplane = np.delete(finite_vox_inplane, inval, 1)
            finite_p2w_normed = np.delete(finite_p2w_normed, inval, 1)
            finite_vox_prods = np.delete(finite_vox_prods, inval)
        ## otherwise, find the distance along the plane to the voxel center...
        finite_vox_inplane -= finite_p2w_normed * finite_vox_prods
        ## the distance to the line is just the norm of this then
        mindist2s = np.sum(finite_vox_inplane ** 2, 0)
        # We now have the in-plane distances to the closest vertices; we can just save this
        d[i, finite_vox_idcs] = mindist2s
    ## At this point we've filled the d-values (those not infinite) to the in-plane squared-
    ## distances from the relevant voxel center to the surface segment; we can turn d into weights;
    ## all infinite distances will automatically get a weight of 0
    d = np.exp(-0.5 * d / (sigma * sigma))
    ## we want to divide weights by the total weight so that all weight columns add up to 1...
    wtot = np.sum(d, 0)
    ### we have to check if any of these values are 0 and give them a nearest-vertex assignment
    zero_idcs = np.where(wtot < 1e-9)
    if len(zero_idcs) > 0:
        (d0s, nei0s) = near.query(idcs.T[zero_idcs], k=1, p=2)
        nei[0, zero_idcs] = nei0s
        d[0, zero_idcs] = 1.0
        d[1:, zero_idcs] = 0.0
        wtot[zero_idcs] = 1.0
    ### okay, now divide
    d /= wtot
    
    # With d = weights and nei = indices, we can now build the dictionary
    vlab = np.array(hemi.white_surface.vertex_labels)
    res = {}
    for i in range(idcs.shape[1]):
        (wh,) = np.where(d[:,i] > 0)
        res[tuple(idcs[:,i].tolist())] = (
            vlab[nei[wh, i]].tolist(),
            d[wh, i].tolist())
    return {[i-1 for i in k]: v for (k,v) in res}

def _line_voxel_overlap(vx, xs, xe):
    '''
    _line_voxel_overlap((i,j,k), xs, xe) yields the fraction of the vector from xs to xe that
      overlaps with the voxel (i,j,k); i.e., the voxel that starts at (i,j,k) and ends at
      (i,j,k) + 1.
    '''
    # We want to consider the line segment from smaller to larger coordinate:
    seg = [(x0,x1) if x0 <= x1 else (x1,x0) for (x0,x1) in zip(xs,xe)]
    # First, figure out intersections of start and end values
    isect = [(min(seg_end, vox_end), max(seg_start, vox_start))
             if seg_end >= vox_start and seg_start <= vox_end else None
             for ((seg_start, seg_end), (vox_start, vox_end)) in zip(seg, [(x,x+1) for x in vx])]
    # If any of these are None, we can't possibly overlap
    if None in isect or any(a == b for (a,b) in isect):
        return 0.0
    return np.linalg.norm([e - s for (s,e) in isect]) \
        / np.linalg.norm([e - s for (s,e) in zip(xs,xe)])

def cortex_to_ribbon_map_lines(sub, hemi=None):
    '''
    cortex_to_ribbon_map_lines(sub) yields a dictionary whose keys are the indices of the ribbon 
      voxels for the given FreeSurfer subject sub and whose values are a tuple of both (0) a list of
      vertex labels associated with the given voxel and (1) a list of the weights associated with
      each vertex (both in identical order).

    These associations are determined by projecting the vectors from the white surface vertices to 
      the pial surface vertices into the the ribbon and weighting them by the fraction of the vector
      that lies in the voxel.

    The following options are accepted:
      * hemi (default: None) specifies which hemisphere to operate over; this may be 'lh', 'rh', or
        None (to do both)
    '''
    
    # we can speed things up slightly by doing left and right hemispheres separately
    if hemi is None:
        return (cortex_to_ribbon_map_lines(sub, hemi='lh'),
                cortex_to_ribbon_map_lines(sub, hemi='rh'))
    if not isinstance(hemi, basestring):
        raise ValueError('hemi must be a string \'lh\', \'rh\', or None')
    if hemi.lower() == 'lh' or hemi.lower() == 'left':
        hemi = sub.LH
    elif hemi.lower() == 'rh' or hemi.lower() == 'right':
        hemi = sub.RH
    else:
        raise RuntimeError('Unrecognized hemisphere: ' + hemi)

    # first we find the indices of all voxels that are in the ribbon
    ribdat = hemi.ribbon.get_data()
    # 1-based indexing is assumed:
    idcs = map(tuple, np.transpose(ribdat.nonzero()) + 1)
    # we also need the transformation from surface to voxel
    tmtx = np.linalg.inv(hemi.ribbon.header.get_vox2ras_tkr())
    # given that the voxels assume 1-based indexing, this means that the center of each voxel is at
    # (i,j,k) for integer (i,j,k) > 0; we want the voxels to start and end at integer values with 
    # respect to the vertex positions, so we subtract 1/2 from the vertex positions, which should
    # make the range 0-1, for example, cover the vertices in the first voxel
    txcoord = lambda mtx: np.dot(tmtx[0:3], np.vstack((mtx, np.ones(mtx.shape[1])))) + 1.5

    # Okay; get the transformed coordinates for white and pial surfaces:
    pialX = txcoord(hemi.pial_surface.coordinates)
    whiteX = txcoord(hemi.white_surface.coordinates)
    
    # make a list of voxels through which each vector passes:
    min_idx = [min(i) for i in idcs]
    max_idx = [max(i) for i in idcs]
    vtx_voxels = [((i,j,k), (id, olap))
                  for (id, (xs, xe)) in enumerate(zip(whiteX.T, pialX.T))
                  for i in range(int(math.floor(min(xs[0], xe[0]))),
                                 int(math.ceil(max(xs[0], xe[0]))))
                  for j in range(int(math.floor(min(xs[1], xe[1]))),
                                 int(math.ceil(max(xs[1], xe[1]))))
                  for k in range(int(math.floor(min(xs[2], xe[2]))),
                                 int(math.ceil(max(xs[2], xe[2]))))
                  for olap in [_line_voxel_overlap((i,j,k), xs, xe)]
                  if olap > 0]
    
    # and accumulate these lists... first group by voxel index then sum across these
    first_fn = lambda x: x[0]
    vox_byidx = {
        vox: ([q[0] for q in dat], [q[1] for q in dat])
        for (vox,xdat) in itertools.groupby(sorted(vtx_voxels, key=first_fn), key=first_fn)
        for dat in [[q[1] for q in xdat]]}
    return {
        tuple([i-1 for i in idx]): (ids, np.array(olaps) / np.sum(olaps))
        for idx in idcs
        if idx in vox_byidx
        for (ids, olaps) in [vox_byidx[idx]]}

def cortex_to_ribbon_map(sub, hemi=None, method='lines', options={}):
    '''
    cortex_to_ribbon_map(sub) yields a dictionary whose keys are the indices of the ribbon voxels
      for the given FreeSurfer subject sub and whose values are a tuple of both (0) a list of vertex
      labels associated with the given voxel and (1) a list of the weights associated with each
      vertex (both in identical order).

    The following options may be given:
      * hemi (default: None) may be 'lh', 'rh', or None; if Nonem then the result is a tuple of the
        (LH_dict, RH_dict) where LH_dict and RH_dict are the individual results from calling the
        cortex_to_ribbon_map function with hemi set to 'lh' or 'rh', respectively
      * method (default: 'lines') may be 'lines' or 'smooth'; the lines method projects the line
        segments from the white to the pial into the volume and weights each voxel by the length of
        the line segments contained in them; the smooth method uses Gaussian weights based on
        distance of the nearby vertices to the voxel center; see cortex_to_ribbon_map_lines and
        cortex_to_ribbon_map_smooth for more details.
      * options (default: {}) may be set to a dictionary of options to pass along to the lines or
        smooth functions
    '''
    if method.lower() == 'lines':
        return cortex_to_ribbon_map_lines(sub, hemi=hemi, **options)
    elif method.lower() == 'smooth':
        return cortex_to_ribbon_map_smooth(sub, hemi=hemi, **options)
    else:
        raise RuntimeError('Unrecognized method: ' + str(method))


def _cortex_to_ribbon_map_into_volume_array(vol, m, dat, method):
    dat = np.array(dat)
    if method == 'weighted':
        for k,v in m.iteritems():
            vol[k[0], k[1], k[2]] = np.dot(dat[v[0]], v[1])
    elif method == 'max':
        for k,v in m.iteritems():
            mx = max(range(len(v[0])), key=lambda i: v[1][i])
            vol[k[0], k[1], k[2]] = dat[v[0][mx]]
    else:
        raise ValueError('Unsupported method %s' % method)
    return vol

def cortex_to_ribbon(sub, data, map='lines', hemi=None, method='weighted',
                     fill=0, dtype=None):
    '''
    cortex_to_ribbon(sub, data) projects the given cortical-surface data to the given subject's
    ribbon and returns the resulting MGHImage volume.
    
    The following options may be given:
      * map (default: 'lines') specifies that a particular cortex_to_ribbon mapping should be used;
        such a mapping can be generated by the cortex_to_ribbon_map function. If None, then the
        mapping is generated. May additionally be a string or a (string, dict) tuple; in this case
        the string is taken to be the method parameter to cortex_to_ribbon_map and the dict is
        taken to be the options.
      * hemi (default: None) specifies that only the given hemisphere should be projected; if this
        is None, then both hemispheres are projected.
      * method (default: 'weighted') specifies the method to use when projecting the map into the
        volume. The value 'weighted' indicates that a weighted mean of vertex data should be used to
        fill each voxel, while the value 'max' indicates that the value of the vertex with the max
        weight should be chosen for each voxel. Generally, 'max' should be used for integer,
        category, or label data.
      * fill (default: 0) specifies the value to be assigned to all voxels outside of the ribbon.
      * dtype (default: None) specifies the data type that should be exported. If None, this will be
        automatically set to np.float32 for floating-point data and np.int32 for integer data.
    '''
    hemi = hemi.lower() if isinstance(hemi, basestring) else hemi
    # First, interpret the arguments:
    mtd = None
    if map is None:
        map = cortex_to_ribbon_map(sub, hemi=hemi, method='lines')
    if isinstance(map, TupleType):
        if len(map) < 1 or len(map) > 2: raise ValueError('Invalid map tuple: %s' % map)
        if isinstance(map[0], basestring):
            (mtd, mtdopts) = map if len(map) == 2 else (map[0], {})
            map = cortex_to_ribbon_map(sub, hemi=hemi, method=mtd, options=mtdopts)
    elif isinstance(map, basestring):
        map = cortex_to_ribbon_map(sub, hemi=hemi, method=map)
    else:
        raise ValueError('Could not interpret map option: %s' % map)
    if hemi is None:
        for (x,xnm) in zip([map, data], ['map', 'data']):
            if not isinstance(x, TupleType) or len(x) != 2:
                raise ValueError('%s must match hemi argument' % xnm)
    if not isinstance(map,  TupleType): map  = (map,  None) if hemi == 'lh' else (None, map)
    if not isinstance(data, TupleType): data = (data, None) if hemi == 'lh' else (None, data)
    # Figure out the dtype
    if dtype is None:
        if method == 'max' and all(isinstance(d, Integral)
                                   for dat in data if dat is not None
                                   for d in dat):
            dtype = np.int32
        else:
            dtype = np.float32
    # start with a duplicate of the left ribbon data
    vol0 = sub.LH.ribbon if hemi is None or hemi == 'lh' else sub.RH.ribbon
    vol0dims = vol0.get_data().shape
    arr = fill * np.ones(vol0dims) if fill != 0 else np.zeros(vol0dims)
    # apply the surface data maps to the volume
    if hemi is None or hemi == 'lh':
        arr = _cortex_to_ribbon_map_into_volume_array(arr, map[0], data[0], method=method)
    if hemi is None or hemi == 'rh':
        arr = _cortex_to_ribbon_map_into_volume_array(arr, map[1], data[1], method=method)
    hdr = vol0.header.copy()
    hdr.set_data_dtype(dtype)
    arr = np.asarray(np.round(arr) if np.issubdtype(dtype, np.int) else arr,
                     dtype=dtype)
    return MGHImage(arr, vol0.affine, hdr, vol0.extra, vol0.file_map)
    
    
    
    
    
